# -*- coding: utf-8 -*-
import os
import random
import platform
from time import sleep
import sys

TYPING_SPEED = 50


def simulate_command_type(commands, shell=False):
    for command in commands.split('&& '):
        if not command.startswith('source'):
            simulate_type(command, shell=shell)


def simulate_type(command, shell=False):
    sys.stdout.write('breno@localhost: ~$ ')
    for c in command:
        sys.stdout.write(c)
        sys.stdout.flush()
        sleep(random.random()*10.0/TYPING_SPEED)
    print('')


# brew install tcl-tk (MAC OS)
# apt-get install tcl python3-tk (UNIX)
def display_subtitle(message, duration=4):
    import tkinter as tk
    root = tk.Tk()
    # Hide the root window drag bar and close button
    root.overrideredirect(True)
    # Make the root window always on top
    root.wm_attributes("-topmost", True)
    # root.wm_attributes("-transparent", True)
    # root.config(bg='systemTransparent')
    root.attributes('-alpha', 0.8)
    root.configure(background='black')
    l = list()
    break_line = False
    for i, letter in enumerate(message):
        if break_line or i and i % 50 == 0:
            if letter == ' ':
                l.append('\n')
                break_line = False
            else:
                break_line = True
        l.append(letter)
    message = ''.join(l)
    line_breaks = message.count('\n')
    if 'darwin' in platform.system().lower():
        font_size, width, top, right = 30, 82, int(root.winfo_screenwidth()/2-20*35), int(root.winfo_screenheight() - [90, 120, 156][line_breaks])
    else:
        font_size, width, top, right = 20, 70, int(root.winfo_screenwidth()/2-70*7.5), int(root.winfo_screenheight() - [77, 110, 140][line_breaks])
    label = tk.Label(root, text=message, font=("Helvetica", font_size), width=width, height=2+line_breaks)
    label.configure(foreground="white", background='black')
    label.pack(expand=tk.YES, fill=tk.BOTH)
    root.geometry("+{}+{}".format(top, right))
    root.after(1000*duration, lambda: root.destroy())
    root.mainloop()


# -*- coding: utf-8 -*-

from djangoplus import cache
from djangoplus.utils.metadata import set_metadata, iterable


def meta(verbose_name, help_text=None, formatter=None, dashboard=None, can_view=()):
    def decorate(func):
        set_metadata(func, 'type', 'attr')
        set_metadata(func, 'verbose_name', verbose_name)
        set_metadata(func, 'can_view', can_view)
        set_metadata(func, 'help_text', help_text)
        set_metadata(func, 'formatter', formatter)
        set_metadata(func, 'dashboard', dashboard)
        return func
    return decorate


def subset(title, help_text=None, list_display=(), list_filter=None, search_fields=None, template=None, menu=None,
           dashboard=None, inline=False, usecase=None, can_view=(), can_alert=(), can_notify=()):
    def decorate(function):
        set_metadata(function, 'type', 'subset')
        set_metadata(function, 'tab', True)
        set_metadata(function, 'title', title)
        set_metadata(function, 'alert', can_alert)
        set_metadata(function, 'notify', can_notify)
        set_metadata(function, 'menu', menu)
        set_metadata(function, 'help_text', help_text)
        set_metadata(function, 'usecase', usecase)
        set_metadata(function, 'can_view', iterable(can_view))
        set_metadata(function, 'inline', inline)
        set_metadata(function, 'name', function.__name__)
        set_metadata(function, 'order', cache.next_number())
        set_metadata(function, 'dashboard', dashboard)
        set_metadata(function, 'list_display', list_display)
        set_metadata(function, 'list_filter', list_filter)
        set_metadata(function, 'search_fields', search_fields)
        set_metadata(function, 'template', template)
        return function

    return decorate


def action(verbose_name, help_text=None, condition=None, inline=(), icon=None, category='Ações', style='popup',
           message='Ação realizada com sucesso.', redirect_to=None, menu=None, initial=None, choices=None, display=None,
           input=None, usecase=None, can_execute=(), can_execute_by_organization=None, can_execute_by_unit=None,
           can_execute_by_role=None):
    def decorate(function):
        function._action = dict(
            title=verbose_name, can_execute=iterable(can_execute), help_text=help_text,
            input=input, group=category or verbose_name, css=style, condition=condition, view_name=function.__name__,
            message=message, initial=initial or '{}_initial'.format(function.__name__), function=function,
            choices=choices or '{}_choices'.format(function.__name__), inline=inline, icon=icon, doc=function.__doc__,
            usecase=usecase, can_execute_by_organization=iterable(can_execute_by_organization),
            can_execute_by_unit=iterable(can_execute_by_unit), can_execute_by_role=iterable(can_execute_by_role),
            redirect_to=redirect_to, menu=menu, display=display
        )
        return function

    return decorate


def role(username, email=None, name=None, active=None, scope=None, signup=False, notify=False):
    def decorate(cls):
        metaclass = getattr(cls, '_meta')
        metaclass.role_username = username
        metaclass.role_email = email
        metaclass.role_name = name
        metaclass.role_active = active
        metaclass.role_scope = scope
        metaclass.role_signup = signup
        metaclass.role_notify = notify
        return cls
    return decorate



import datetime

from djangoplus.utils.permissions import has_add_permission

from djangoplus.ui import RequestComponent
from djangoplus.utils.metadata import get_metadata
from djangoplus.utils.serialization import dumps_qs_query


class Calendar(RequestComponent):
    def __init__(self, request, title, url=None):
        super(Calendar, self).__init__(title, request)
        self.title = title
        self.items = []
        self.lazy = False
        self.initial_date = None
        self.url = url

        self.set_initial_date(datetime.date.today())

    def add(self, description, start, end=None, url=None, allday=True):
        item = dict(description=description, start=start, end=end, allday=allday)
        self.items.append(item)
    
    def set_initial_date(self, initial_date):
        self.initial_date = initial_date
        session_initial_date = self.request.session.get('calendar_initial_date')
        if session_initial_date:
            self.initial_date = datetime.datetime.strptime(session_initial_date, '%Y-%m-%d')


class ModelCalendar(Calendar):
    def __init__(self, request, title, editable=False, display_time=True, url=None):
        super(ModelCalendar, self).__init__(request, title, url=url)
        self.links = []
        self.lazy = True
        self.models = []
        self.editable = editable
        self.display_time = display_time

    def add(self, queryset, start_field, end_field=None, color='#ccc', action_names=[], as_initial_date=False):
        item = dict(queryset=dumps_qs_query(queryset), start_field=start_field, end_field=end_field, color=color, action_names=','.join(action_names))
        self.items.append(item)
        if as_initial_date:
            qs_initial_date = queryset.order_by(start_field).values_list(start_field, flat=True)
            if qs_initial_date.exists():
                self.set_initial_date(qs_initial_date[0])
        if queryset.model not in self.models and has_add_permission(self.request, queryset.model):
            label = get_metadata(queryset.model, 'verbose_name')
            app_label = get_metadata(queryset.model, 'app_label')
            model_name = queryset.model.__name__.lower()
            url = '/add/{}/{}/?{}='.format(app_label, model_name, start_field)
            link = dict(label=label, url=url)
            self.links.append(link)
            self.models.append(queryset.model)


# -*- coding: utf-8 -*-
from django.core.management.base import BaseCommand

from djangoplus.tutorial.utils import display_subtitle


class Command(BaseCommand):

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('words', nargs='*', default=None)

    def handle(self, *args, **options):
        words = options.pop('words')
        display_subtitle(' '.join(words).replace('\\n', '\n'))



# -*- coding: utf-8 -*-
import os
from djangoplus.utils.webdriver import Browser
from djangoplus.tutorial.tools import VideoRecorder, HttpServer, Terminal, Subtitle


VIRTUALENV_PATH = '/var/opt/.virtualenvs/djangoplus3'
VIRTUALENV_ACTIVATION_PATH = '{}/bin/activate'.format(VIRTUALENV_PATH)
PYTHON_PATH = '{}/bin/python'.format(VIRTUALENV_PATH)


class Tutorial(object):
    def __init__(self, project_path, devmode=False):

        self.project_path = project_path
        self.project_name = self.project_path.split(os.sep)[-1]
        self.base_dir = '/tmp/{}'.format(self.project_name)
        self.devmode = devmode

        self.subtitle = Subtitle()
        self.recorder = VideoRecorder()
        self.terminal = Terminal(python=PYTHON_PATH)
        self.server = HttpServer(self.base_dir, devmode, python=PYTHON_PATH)
        self.browser = Browser('http://localhost:8000', verbose=False, slowly=True, maximize=True)
        self.browser.get("https://www.google.com.br/")

    def start(self, create_project=True):
        try:
            self.terminal.hide()
            if create_project:
                self.delete_project()
                self.create_project()
            self.server.start()
            self.start_tutorial()
        finally:
            self.recorder.stop()
            self.server.stop()
            self.terminal.show()
            self.browser.close(5)

    def create_project(self):
        if self.devmode:
            os.system('clear')
        os.environ['DJANGO_SETTINGS_MODULE'] = '{}.settings'.format(self.project_name)
        command = 'source {} && cd /tmp && startproject {}'.format(VIRTUALENV_ACTIVATION_PATH, self.project_name)
        self.terminal.execute(command)

    def delete_project(self):
        os.system('rm -r {}'.format(self.base_dir))

    def start_tutorial(self):
        pass


class DevTutorial(Tutorial):
    def __init__(self, project_path):
        super(DevTutorial, self).__init__(project_path, devmode=True)

    def restart_server(self, sync=False):
        self.server.stop()
        if sync:
            self.terminal.execute('python manage.py sync', base_dir=self.base_dir)
        self.server.start()

    def edit_models(self, step, file_name='models.py'):
        input_file_path = '{}/{}/{}'.format(self.project_path, self.project_name, file_name)
        output_file_path = '{}/{}/{}'.format(self.base_dir, self.project_name, file_name)
        print('Reading file {}'.format(input_file_path))
        input_file = open(input_file_path, 'r')
        file_content = input_file.read()
        input_file.close()
        print('Writing file {}'.format(output_file_path))
        output_file = open(output_file_path, 'w')
        output_file.write(file_content)
        output_file.close()
        self.restart_server(True)

    def edit_templates(self, step, file_name):
        pass

    def edit_formatters(self, file_name='formatters.py'):
        pass


class UserTutorial(Tutorial):
    def __init__(self, project_path):
        super(UserTutorial, self).__init__(project_path, devmode=False)



# -*- coding: utf-8 -*-
import datetime
from decimal import Decimal
from django import template
from django.utils.safestring import mark_safe
from django.template.loader import render_to_string
from djangoplus.utils import permissions
from djangoplus.utils.metadata import get_metadata
from django.template.defaultfilters import slugify
register = template.Library()


@register.simple_tag()
def tree_info(obj, queryset):
    if hasattr(obj, 'get_parent_field'):
        parent_field = obj.get_parent_field()
        if parent_field:
            parent = getattr(obj, parent_field.name)
            if not hasattr(queryset, '__pks'):
                queryset.__pks = queryset.values_list('pk', flat=True)
            if parent and parent.pk in queryset.__pks:
                return 'treegrid-{}  treegrid-parent-{}'.format(obj.pk, parent.pk)
            else:
                return 'treegrid-{}'.format(obj.pk)
    else:
        return ''


@register.simple_tag()
def paginator_checkboxes(paginator, obj, as_row=False):
    l = []
    if paginator.display_checkboxes:
        if as_row:
            l.append('<td style="vertical-align: middle" width="5px">')
        if obj:
            l.append('<label class="label-checkbox"><input type="checkbox" name="pk" value="{}" onclick="check{}();"><span class="custom-checkbox"></span></label>'.format(obj.pk, paginator.id))
        else:
            l.append("""<label class="label-checkbox"><input name="pk" type="checkbox" value="0" onclick="$('input[name=\\\'pk\\\']').prop('checked', this.checked);check{}();"><span class="custom-checkbox"></span></label>""".format(paginator.id))
        if as_row:
            l.append('</td>')
    return mark_safe(''.join(l))


@register.simple_tag()
def paginator_icons(paginator, obj, as_button=False):
    relation = paginator.relation
    edit = not paginator.readonly
    delete = not paginator.readonly
    return obj_icons(paginator.request, obj, relation=relation, edit=edit, delete=delete, as_button=as_button)


@register.simple_tag()
def obj_icons(request, obj, relation=None, edit=True, delete=True, css=None, as_button=False):
    l = []

    if as_button:
        css = 'btn btn-default {}'.format(css or '{}')
    else:
        css = 'hide-label {}'.format(css or '{}')

    if relation:
        view_url = relation.view_url.format(obj.pk)
        btn = '<a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-search fa-lg"></i><span> Visualizar</span></a>'
        l.append(btn.format(slugify(view_url), css.format('ajax'), view_url, 'Visualizar'))

        if relation.edit_url:
            if relation.is_one_to_many or relation.is_many_to_many:
                has_edit_permission = permissions.has_edit_permission(request, relation.model)
            else:
                has_edit_permission = permissions.has_edit_permission(request, relation.relation_model)
            if has_edit_permission and (not hasattr(obj, 'can_edit') or obj.can_edit()):
                edit_url = relation.edit_url.format(obj.pk)
                btn = ' <a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-edit fa-lg"></i><span> Editar</span></a>'
                l.append(btn.format(slugify(edit_url), css.format('popup'), edit_url, 'Editar'))

        if relation.delete_url:
            if relation.is_one_to_many or relation.is_many_to_many:
                has_delete_permission = delete
            else:
                has_delete_permission = permissions.has_delete_permission(request, relation.relation_model)
            if has_delete_permission and (not hasattr(obj, 'can_delete') or obj.can_delete()):
                delete_url = relation.delete_url.format(obj.pk)
                btn = ' <a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-close fa-lg"></i><span> Excluir</span></a>'
                l.append(btn.format(slugify(delete_url), css.format('popup'), delete_url, 'Excluir'))
    else:
        model = type(obj)
        cls = model.__name__.lower()
        app = get_metadata(model, 'app_label')

        tree_index_field = None
        if hasattr(obj, 'get_tree_index_field'):
            tree_index_field = obj.get_tree_index_field()

        view_url = hasattr(obj, 'get_absolute_url') and obj.get_absolute_url() or '/view/{}/{}/{}/'.format(app, cls, obj.pk)
        btn = '<a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-search fa-lg"></i><span> Visualizar</span></a>'
        l.append(btn.format(slugify(view_url), css.format('ajax'), view_url, 'Visualizar'))

        if edit and permissions.has_edit_permission(request, model) and (not hasattr(obj, 'can_edit') or obj.can_edit()):
            edit_url = '/add/{}/{}/{}/'.format(app, cls, obj.pk)
            btn = ' <a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-edit fa-lg"></i><span> Editar</span></a>'
            l.append(btn.format(slugify(edit_url), css.format('ajax'), edit_url, 'Editar'))

        if delete and permissions.has_delete_permission(request, model) and (not hasattr(obj, 'can_delete') or obj.can_delete()):
            delete_url = '/delete/{}/{}/{}/'.format(app, cls, obj.pk)
            btn = ' <a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-close fa-lg"></i><span> Excluir</span></a>'
            l.append(btn.format(slugify(delete_url), css.format('popup'), delete_url, 'Excluir'))

        if tree_index_field:
            add_url = '/add/{}/{}/{}/{}/'.format(app, cls, obj.pk, cls)
            btn = ' <a id="{}" class="{}" href="{}" title="{}"><i class="fa fa-plus fa-lg"><span> Adicionar</span></i></a>'
            l.append(btn.format(slugify(view_url), css.format('popup'), add_url, 'Adicionar'))

    return mark_safe(''.join(l))


@register.filter()
def align(value):
    position ='left'
    if isinstance(value, bool) or isinstance(value, datetime.date):
        position='center'
    elif isinstance(value, Decimal) or isinstance(value, int):
        position='right'
    elif isinstance(value, tuple):
        return align(value[0])
    return position


@register.filter
def column_name(paginator, i):
    return paginator.column_names[i][0]


@register.filter
def paginate(paginator):
    return render_to_string(paginator.template, {'self': paginator})


@register.simple_tag()
def add_grouped_actions(paginator, obj):
    return add_actions(paginator, obj, category='Ações')


@register.simple_tag()
def add_actions(paginator, obj, category=None):
    paginator.drop_down.add_actions(obj, inline=True, subset_name=paginator.get_current_tab_name() or None, category=category)
    return ''


# -*- coding: utf-8 -*-

from django.conf import settings
from djangoplus.ui.components import forms
from django.apps import apps
from djangoplus.utils.metadata import get_metadata, find_field_by_name, list_related_objects, get_fiendly_name, getattr2


def get_register_form(request, obj):

    _model = type(obj)

    initial = hasattr(obj, 'initial') and obj.initial() or {}
    choices = hasattr(obj, 'choices') and obj.choices() or {}

    form_name = get_metadata(_model, 'add_form')
    app_label = get_metadata(_model, 'app_label')
    verbose_name = get_metadata(_model, 'verbose_name')
    role_username = get_metadata(_model, 'role_username', None)

    if form_name:
        full_app_name = settings.APP_MAPPING.get(app_label, app_label)
        forms_module = __import__('{}.forms'.format(full_app_name), fromlist=list(map(str, [app_label])))
        Form = getattr(forms_module, form_name)
    else:
        if obj.pk:
            form_title = 'Atualização de {}'.format(str(verbose_name))
            button_label = 'Atualizar'
        else:
            add_label = get_metadata(_model, 'add_label', None)
            form_title = add_label or 'Cadastro de {}'.format(str(verbose_name))
            button_label = add_label or 'Cadastrar'

        class Form(forms.ModelForm):
            class Meta:
                model = _model
                fields = get_metadata(_model, 'form_fields', '__all__')
                exclude = get_metadata(_model, 'exclude_fields', ())
                submit_label = button_label
                title = form_title
                icon = get_metadata(_model, 'icon', None)
                perm_or_group = '{}.add_{}'.format(app_label, _model.__name__.lower())

    form = Form(request, instance=obj, initial=initial)
    form.name = '{}Form'.format(_model.__name__)
    for field_name in choices:
        form.fields[field_name].queryset = choices[field_name]
        form.fields[field_name].widget.queryset = choices[field_name]

    if hasattr(obj, 'get_parent_field'):
        parent_field = obj.get_parent_field()
        if parent_field:
            if not obj.pk and parent_field.name in form.fields:
                form.fields[parent_field.name].widget = forms.widgets.HiddenInput()

    if role_username and obj.pk:
        if role_username in form.fields:
            form.fields[role_username].widget = forms.widgets.ReadOnlyInput()

    return form


def get_one_to_many_form(request, obj, related_field_name, **kwargs):
    _model = type(obj)
    related_field = find_field_by_name(_model, related_field_name)

    class Form(forms.ModelForm):
        class Meta:
            model = related_field.remote_field.model
            fields = get_metadata(related_field.remote_field.model, 'form_fields', '__all__')
            exclude = get_metadata(related_field.remote_field.model, 'exclude_fields', ())
            submit_label = 'Adicionar {}'.format(get_metadata(related_field.remote_field.model, 'verbose_name'))
            title = 'Adicionar {}'.format(get_metadata(related_field.remote_field.model, 'verbose_name'))
            icon = get_metadata(related_field.remote_field.model, 'icon', None)
            is_inner = True

        def save(self, *args, **kwargs):
            super(Form, self).save(*args, **kwargs)
            getattr(obj, related_field_name).add(self.instance)

    form = Form(request, **kwargs)
    return form


def get_many_to_one_form(request, obj, related_field_name, related_obj):
    _model = type(obj)
    rel = getattr(_model, related_field_name).rel
    add_label = get_metadata(rel.related_model, 'add_label')
    if add_label:
        form_title = add_label
        button_label = add_label
    else:
        action = related_obj.pk and 'Atualizar' or 'Adicionar'
        form_title = '{} {}'.format(action, get_metadata(rel.related_model, 'verbose_name'))
        button_label = action
    related_field_name = rel.field.name

    setattr(related_obj, related_field_name, obj)

    initial = hasattr(related_obj, 'initial') and related_obj.initial() or {}
    choices = hasattr(related_obj, 'choices') and related_obj.choices() or {}

    app_label = get_metadata(_model, 'app_label')

    form_name = get_metadata(rel.related_model, 'add_form')
    if form_name:
        full_app_name = settings.APP_MAPPING.get(app_label, app_label)
        forms_module = __import__('{}.forms'.format(full_app_name), fromlist=list(map(str, [app_label])))
        Form = getattr(forms_module, form_name)
    else:
        class Form(forms.ModelForm):
            class Meta:
                model = rel.related_model
                fields = get_metadata(rel.related_model, 'form_fields', '__all__')
                exclude = get_metadata(rel.related_model, 'exclude_fields', ())
                submit_label = related_obj.pk and 'Atualizar' or button_label
                title = form_title
                is_inner = True

    initial[related_field_name] = obj.pk
    for key in list(initial.keys()):
        if hasattr(obj, key) and obj.pk and getattr(obj, key):
            del (initial[key])

    form = Form(request, initial=initial, instance=related_obj)
    form.form_name = '{}Form'.format(rel.related_model.__name__)
    for field_name in choices:
        if field_name in form.fields:
            form.fields[field_name].queryset = choices[field_name]
    if related_field_name in form.fields:
        del (form.fields[related_field_name])
    return form


def get_one_to_one_form(request, obj, related_field_name, related_pk, **kwargs):
    _model = type(obj)

    related_field = find_field_by_name(_model, related_field_name)
    related_object = related_pk and related_field.remote_field.model.objects.get(pk=related_pk)

    initial = hasattr(related_field.remote_field.model, 'initial') and related_field.remote_field.model.initial() or {}
    choices = hasattr(related_field.remote_field.model, 'choices') and related_field.remote_field.model.choices() or {}

    class Form(forms.ModelForm):
        class Meta:
            model = related_field.remote_field.model
            fields = get_metadata(related_field.remote_field.model, 'form_fields', '__all__')
            exclude = get_metadata(related_field.remote_field.model, 'exclude_fields', ())
            submit_label = 'Atualizar {}'.format(related_field.verbose_name)
            title = 'Atualizar {}'.format(related_field.verbose_name)
            icon = get_metadata(related_field.remote_field.model, 'icon', None)
            is_inner = True

        def save(self, *args, **kwargs):
            super(Form, self).save(*args, **kwargs)
            setattr(obj, related_field_name, self.instance)
            _model.objects.filter(pk=obj.pk).update(**{related_field_name: self.instance})

    form = Form(request, instance=related_object, initial=initial, **kwargs)
    form.name = '{}Form'.format(related_field.remote_field.model.__name__)
    for field_name in choices:
        if field_name in form.fields:
            form.fields[field_name].queryset = choices[field_name]
    return form


def get_many_to_many_form(request, obj, related_field_name, related_pk):
    _model = type(obj)

    initial = hasattr(obj, 'initial') and obj.initial() or {}
    choices = hasattr(obj, 'choices') and obj.choices() or {}

    related_field_model = find_field_by_name(_model, related_field_name).remote_field.model

    class Form(forms.ModelForm):
        related_objects = forms.MultipleModelChoiceField(related_field_model.objects.all(),
                                                         label=get_metadata(related_field_model, 'verbose_name'))

        class Meta:
            model = _model
            fields = ()
            title = 'Adicionar {}'.format(get_metadata(related_field_model, 'verbose_name'))
            icon = get_metadata(related_field_model, 'icon', None)
            # is_inner = True

        def save(self, *args, **kwargs):
            for related_object in self.cleaned_data['related_objects']:
                getattr(self.instance, related_field_name).add(related_object)

    form = Form(request, instance=obj, initial=initial)
    form.name = '{}Form'.format(_model.__name__)
    for field_name in choices:
        if field_name in form.fields:
            form.fields[field_name].queryset = choices[field_name]
    return form


def get_many_to_many_reverse_form(request, obj, related_field_name):
    _model = type(obj)

    initial = hasattr(obj, 'initial') and obj.initial() or {}
    choices = hasattr(obj, 'choices') and obj.choices() or {}
    field = getattr(_model, '{}_set'.format(related_field_name))
    related_field_model = field.rel.remote_field.model

    class Form(forms.ModelForm):
        related_objects = forms.MultipleModelChoiceField(related_field_model.objects.all(),
                                                         label=get_metadata(related_field_model, 'verbose_name'))

        class Meta:
            model = _model
            fields = ()
            title = 'Adicionar {}'.format(get_metadata(related_field_model, 'verbose_name'))
            icon = get_metadata(related_field_model, 'icon', None)
            # is_inner = True

        def save(self, *args, **kwargs):
            for related_object in self.cleaned_data['related_objects']:
                getattr(related_object, field.rel.remote_field.name).add(self.instance)
            self.instance._check_role()

    form = Form(request, instance=obj, initial=initial)
    form.name = '{}Form'.format(_model.__name__)
    for field_name in choices:
        if field_name in form.fields:
            form.fields[field_name].queryset = choices[field_name]
    return form


def get_class_action_form(request, _model, action, func):
    action_title = action['title']
    initial = action['initial']
    action_input = action['input']
    app_label = get_metadata(_model, 'app_label')

    if action_input:
        # it is a form name
        if type(action_input) in [str, str] and '.' not in action_input:
            full_app_name = settings.APP_MAPPING.get(app_label, app_label)
            module = __import__('{}.forms'.format(full_app_name), fromlist=list(map(str, [app_label])))
            Form = getattr(module, action_input)
        # it is a model or model name
        else:
            if type(action_input) in [str, str]:
                app_name, class_name = action_input.split('.')
                action_input = apps.get_model(app_name, class_name)

            class Form(forms.ModelForm):
                class Meta:
                    model = action_input
                    fields = func.__code__.co_varnames[1:func.__code__.co_argcount]
                    title = action_title
                    submit_label = action_title
    else:
        class Form(forms.ModelForm):
            class Meta:
                model = _model
                fields = func.__code__.co_varnames[1:func.__code__.co_argcount]
                title = action_title
                submit_label = action_title

    initial = hasattr(_model.objects, initial) and getattr(_model.objects, initial)() or None
    form = Form(request, initial=initial)
    return form


def get_action_form(request, obj, action):
    action_function = action['function']
    action_title = action['title']
    initial = action['initial']
    action_input = action['input']
    action_choices = action['choices']
    action_display = action['display']
    app_label = get_metadata(type(obj), 'app_label')
    func = getattr(obj, action_function.__name__, action_function)

    if initial and hasattr(obj, initial):
        initial = getattr(obj, initial)()
    else:
        initial = {}
    if action_choices and hasattr(obj, action_choices):
        action_choices = getattr(obj, action_choices)()
    else:
        action_choices = {}

    if action_input:
        # it is a form name
        if type(action_input) in [str, str] and '.' not in action_input:
            full_app_name = settings.APP_MAPPING.get(app_label, app_label)
            fromlist = app_label
            module = __import__('{}.forms'.format(full_app_name), fromlist=list(map(str, [app_label])))
            form_cls = getattr(module, action_input)

        # it is a model or model name
        else:
            if type(action_input) in [str, str]:
                app_name, class_name = action_input.split('.')
                action_input = apps.get_model(app_name, class_name)

            class Form(forms.ModelForm):
                class Meta:
                    model = action_input
                    fields = func.__code__.co_varnames[1:func.__code__.co_argcount]
                    title = action_title
                    submit_label = action_title

            form_cls = Form
    else:

        class Form(forms.ModelForm):
            class Meta:
                model = func.__self__.__class__
                fields = func.__code__.co_varnames[1:func.__code__.co_argcount]
                title = action_title
                submit_label = action_title

        form_cls = Form

    if issubclass(form_cls, forms.ModelForm):
        for key in list(initial.keys()):
            if hasattr(obj, key) and obj.pk and getattr(obj, key):
                del (initial[key])
        form = form_cls(request, instance=obj, initial=initial)
    else:
        form = form_cls(request, initial=initial)

    if action_display:
        for lookup in action_display:
            label = get_fiendly_name(func.__self__.__class__, lookup)
            value = getattr2(obj, lookup)
            form.fields[lookup] = forms.CharField(label=label, initial=value, required=False, widget=forms.widgets.DisplayInput(value))

    if action_choices:
        for field_name in action_choices:
            form.fields[field_name].queryset = action_choices[field_name]

    if not obj.pk:
        verbose_name = get_metadata(obj.__class__, 'verbose_name')
        form.fields['instance'] = forms.ModelChoiceField(type(obj).objects.all(), label=verbose_name)
        if form.fieldsets:
            form.fieldsets = ((verbose_name, {'fields': ('instance',)}),) + form.fieldsets

    return form


def get_delete_form(request, obj):
    class Form(forms.Form):
        class Meta:
            submit_label = 'Confirmar Exclusão'
            submit_style = 'danger'

        def __init__(self, *args, **kwargs):
            super(Form, self).__init__(*args, **kwargs)
            self.related_objects = []
            for related_object in list_related_objects(type(obj)):
                try:
                    qs = getattr(obj, related_object.get_accessor_name()).all()
                except:
                    continue
                verbose_name = get_metadata(qs.model, 'verbose_name_plural')
                count = qs.count()
                if count:
                    self.related_objects.append(dict(verbose_name=verbose_name, count=count))

    form = Form(request)
    return form



# -*- coding: utf-8 -*-
import copy
from collections import OrderedDict
from datetime import datetime
from djangoplus.ui import RequestComponent
from djangoplus.utils import permissions
from djangoplus.utils.metadata import get_metadata, get_can_execute
from djangoplus.utils.metadata import check_condition
from djangoplus.utils import http


class GroupDropDown(RequestComponent):
    def __init__(self, request):
        super(GroupDropDown, self).__init__('groupdropdown', request)
        self.actions = OrderedDict()
        self.mobile = http.mobile(self.request)
        self.actions['Ações'] = []

    def add_action(self, label, url, css='popup', icon=None, category='Ações'):
        if self.mobile:
            category = 'Ações'
        if category not in self.actions:
            self.actions[category] = []
        item = dict(label=label, url=url, css=css, icon=icon)
        self.actions[category].append(item)

    def has_items(self):
        for category in self.actions:
            if self.actions[category]:
                return True
        return False


class ModelDropDown(GroupDropDown):
    def __init__(self, request, model, action_names=()):
        from djangoplus.cache import loader
        self.model = model
        self.action_names = action_names
        self.actions_cache = None
        self.hash = abs(datetime.now().__hash__())
        self.obj = None

        self.has_inline_action = False
        if action_names and len(action_names) > 0:
            self.has_inline_action = True
        else:

            for category in loader.actions[self.model]:
                for view_name in loader.actions[self.model][category]:
                    form_action = loader.actions[self.model][category][view_name]
                    if not self.has_inline_action:
                        self.has_inline_action = form_action.get('inline')

            if not self.has_inline_action:
                self.has_inline_action = self.model in loader.add_inline_actions

        super(ModelDropDown, self).__init__(request)
        # adding the actions defined in the model class
        for category in loader.actions[self.model]:
            if category not in self.actions:
                self.actions[category] = []

    def add_action(self, label, url, css='popup', icon=None, category='Ações'):
        if category not in self.actions:
            self.actions[category] = []
        if self.obj:
            if '{}' in url:
                url = url.format(self.obj.pk)
            else:
                url = '{}{}/'.format(url, self.obj.pk)
        super(ModelDropDown, self).add_action(label, url, css, icon, category)

    def add_actions(self, obj, inline=False, fieldset_title=None, subset_name=None, category=None):
        from djangoplus.cache import loader
        obj.request = self.request
        if not self.actions_cache:
            self.actions_cache = self.actions

        self.hash = abs(datetime.now().__hash__())
        self.actions = copy.deepcopy(self.actions_cache)
        self.obj = obj

        if inline:
            if self.model in loader.add_inline_actions:
                for add_inline_action in loader.add_inline_actions[self.model]:
                    if add_inline_action['subset'] is True or add_inline_action['subset'] == subset_name:
                        if permissions.check_group_or_permission(self.request, add_inline_action['can_execute']):
                            self.add_action(
                                add_inline_action['title'], add_inline_action['url'], 'popup', 'fa fa-plus'
                            )

        for action_category in loader.actions[self.model]:
            for view_name in loader.actions[self.model][action_category]:
                action = loader.actions[self.model][action_category][view_name]
                action_function = action.get('function')
                action_title = action['title']
                action_can_execute = get_can_execute(action)
                action_condition = action['condition']
                action_css = action['css']
                action_input = action['input']
                action_inline = action['inline']
                action_icon = action['icon']
                action_display = action['display']

                action_name = action_function.__name__
                is_action_view = not hasattr(self.model, action_name)

                if action_name and action_css == 'popup' and not is_action_view:
                    func = getattr(self.model, action_name)
                    action_css = (func.__code__.co_argcount > 1 or action_input or action_display) and action_css or 'ajax'

                # it is a dropdown in a model panel
                if fieldset_title is not None:
                    if fieldset_title:
                        if action_name not in loader.fieldset_actions[self.model][fieldset_title]:
                            continue
                    else:
                        # if the action was included in any fieldset it can not be displayed in page's action panel
                        add_action = True
                        for fieldset_title2, action_names in list(loader.fieldset_actions[self.model].items()):
                            if action_name in action_names:
                                add_action = False
                                break
                        if not add_action:
                            continue
                else:
                    # it is a dropdown in a paginator
                    if action_inline is not True and (subset_name not in loader.subset_actions[self.model] or action_name not in loader.subset_actions[self.model][subset_name]):
                        continue

                if not permissions.check_group_or_permission(self.request, action_can_execute):
                    continue

                if not check_condition(action_condition, obj):
                    continue

                if is_action_view:
                    action_url = '/{}/{}/'.format(get_metadata(self.model, 'app_label'), action_name)
                else:
                    action_url = '/action/{}/{}/{}/'.format(get_metadata(self.model, 'app_label'), self.model.__name__.lower(), view_name)

                self.add_action(action_title, action_url, action_css, action_icon, category or action_category)



# -*- coding: utf-8 -*-
import uuid
from django.utils.text import slugify
from django.utils.safestring import mark_safe
from djangoplus.utils.http import ReportResponse
from django.utils.deprecation import MiddlewareMixin
from django.template.loader import render_to_string


class Component(object):

    class Media:
        css = {'all': ()}
        js = ()

    def __init__(self, uid=None):
        self.id = slugify(str(uid or uuid.uuid1().hex))
        self.as_pdf = False

    def __str__(self):
        return self.render('{}.html'.format(self.__class__.__name__.lower()))

    def render(self, template_name):
        medias = []
        media_cls = getattr(self, 'Media')
        if hasattr(media_cls, 'js'):
            for script in media_cls.js:
                medias.append('<script src="{}"></script>'.format(script))
        if hasattr(media_cls, 'css'):
            for css_media, urls in media_cls.css.items():
                for url in urls:
                    script = '$("<link/>", {{rel: "stylesheet",type: "text/css",href: "{}"}}).appendTo("head");'.format(url)
                    medias.append('<script>{}</script>'.format(script))
        html = render_to_string(template_name, {'self': self})
        return mark_safe('{}\n{}'.format(''.join(medias), html))


class RequestComponent(Component):

    def __init__(self, uid, request):
        self.title = None
        self.request = request
        self.response = None

        super(RequestComponent, self).__init__(uid)

        if self.request:
            if 'pdf' in self.request.GET:
                self.as_pdf = True

    def process_request(self):
        if self.request:
            if self.request.GET.get('pdf') == self.id:
                raise ComponentHasResponseException(ReportResponse(self.title, self.request, [self]))


class ComponentHasResponseException(Exception):

    def __init__(self, response):
        self.response = response
        super(ComponentHasResponseException, self).__init__()


class ComponenteResponseMiddleware(MiddlewareMixin):

    def process_exception(self, request, exception):
        if type(exception) == ComponentHasResponseException:
            return exception.response
        return None



# -*- coding: utf-8 -*-

from django import template
from djangoplus.ui.components.navigation.breadcrumbs import Breadcrumbs

register = template.Library()


@register.simple_tag()
def breadcrumbs(request, view_title):
    return str(Breadcrumbs(request, view_title))



# -*- coding: utf-8 -*-
import json
from djangoplus.docs.doc import Documentation
from os import path, listdir
from django.conf import settings
from djangoplus.cache import loader
from djangoplus.admin.models import Group
from djangoplus.decorators.views import view, action
from djangoplus.admin.models import User, Organization


@view('Modelo', login_required=False, template='source.html')
def model(request):
    src = []
    content = open(path.join(settings.BASE_DIR, settings.PROJECT_NAME, 'models.py')).read()
    src.append(dict(file='models.py', language='python', content=content))
    content = open(path.join(settings.BASE_DIR, settings.PROJECT_NAME, 'formatters.py')).read()
    src.append(dict(file='formatters.py', language='python', content=content))
    for template_file_name in listdir(path.join(settings.BASE_DIR, settings.PROJECT_NAME, 'templates')):
        if template_file_name != 'public.html':
            content = open(path.join(settings.BASE_DIR, settings.PROJECT_NAME, 'templates', template_file_name)).read()
            src.append(dict(file=template_file_name, language='html', content=content))
    return locals()


@view('Testes', login_required=False, template='source.html')
def tests(request):
    src = []
    content = open(path.join(settings.BASE_DIR, settings.PROJECT_NAME, 'tests.py')).read()
    src.append(dict(file='tests.py', language='python', content=content))
    return locals()


@view('Homologação', login_required=False)
def homologate(request):
    title = 'Homologação'
    groups = Group.objects.filter(role__scope__isnull=True).order_by('name').distinct()

    http_host = request.META['HTTP_HOST']
    if loader.organization_model:
        organization_model_name = loader.organization_model.__name__.lower()
    if loader.unit_model:
        unit_model_name = loader.unit_model.__name__.lower()

    organization_group_names = []
    unit_group_names = []
    for model in loader.role_models:
        name = loader.role_models[model]['name']
        scope = loader.role_models[model]['scope']
        if scope == 'organization':
            organization_group_names.append(name)
        if scope == 'unit':
            unit_group_names.append(name)

    organization_groups = Group.objects.filter(name__in=organization_group_names).order_by('name').distinct()
    unit_groups = Group.objects.filter(name__in=unit_group_names).order_by('name').distinct()

    organizations = []
    for organization in Organization.objects.all():

        organization.users = []
        for group in organization_groups:
            organization.users.append(User.objects.filter(role__group=group, role__organizations=organization))

        organization.units = []
        for unit in organization.get_units():
            organization.units.append(unit)
            unit.users = []
            for group in unit_groups:
                unit.users.append(User.objects.filter(role__group=group, role__units=unit))

        organizations.append(organization)
    return locals()


@view('Doc', login_required=False)
def doc(request):
    documentation = Documentation()
    workflow_data = json.dumps(loader.workflows)
    return locals()


@action(User, 'Login as User', can_execute='Superuser')
def login_as(request, pk):
    pass


# -*- coding: utf-8 -*-

from djangoplus.ui.components.select.widgets import *


class FormattedTextarea(widgets.Textarea):

    class Media:
        js = ('/static/js/tinymce.min.js',)

    def render(self, name, value, attrs={}):
        attrs['class'] = 'form-control'
        html = super(FormattedTextarea, self).render(name, value, attrs)
        js = '''
            <script>
            tinymce.remove();
            tinymce.init({{
              selector: '#id_{}',
              height: 300,
              menubar: false,
                plugins: [
                    'advlist autolink lists link image table textcolor',
                  ],
              toolbar: 'insert | undo redo |  formatselect | bold italic backcolor forecolor  | alignleft aligncenter alignright alignjustify | bullist numlist outdent indent | removeformat | table',
                  content_css: []
            }});
            </script>
        '''.format(name)
        return mark_safe('{}{}'.format(html, js))




# -*- coding: utf-8 -*-
import tempfile
import qrcode
import base64
from decimal import Decimal
from django.conf import settings
from django.http import HttpResponse
from djangoplus.ui.components import forms
from djangoplus.utils.http import ComponentResponse
from djangoplus.ui import RequestComponent, ComponentHasResponseException
from djangoplus.utils.metadata import get_fiendly_name, get_field, get_metadata, getattr2


PERCENT_SYMBOL = '%'
DOLLAR_SYMBOL = '$'
REAIS_SYMBOL = 'R$'


class Chart(RequestComponent):
    def __init__(self, request, labels, series, groups=[], symbol=None, title=None):

        super(Chart, self).__init__(None, request)

        self.type = 'bar'
        self.labels = labels
        self.series = series
        self.groups = groups
        self.symbol = symbol or ''
        self.title = title
        self.color_index = 0
        self.display = False
        self.clickable = False

    def __str__(self):
        for i, serie in enumerate(self.series):
            for j, valor in enumerate(serie):
                if type(valor) == Decimal:
                    self.series[i][j] = float(valor)
                if valor:
                    self.display = True

        self.colors = []
        for color in settings.GRADIENT:
            self.colors.append(color)
        for i in range(len(self.colors), 16):
            self.colors.append('#FFFFFF')

        return super(Chart, self).__str__()

    def get_box_series(self):
        l = []
        serie = self.series[0]
        total = sum(serie)
        for i, label in enumerate(self.labels):
            percentage = int(total and serie[i]*100/total or 0)
            l.append((label, serie[i], percentage))
        return l

    def next_color(self):
        color = self.colors[self.color_index]
        self.color_index += 1
        return color

    def change_type(self, chart_type):
        self.type = chart_type
        return self

    def pie(self):
        return self.change_type('pie')

    def donut(self):
        return self.change_type('donut')

    def bar(self):
        return self.change_type('bar')

    def horizontal_bar(self):
        return self.change_type('horizontalbar')

    def stack(self):
        return self.change_type('stack')

    def horizontal_stack(self):
        return self.change_type('horizontalstack')

    def line(self):
        return self.change_type('line')

    def area(self):
        return self.change_type('area')

    def box(self):
        return self.change_type('box')


class QueryStatisticsChart(Chart):

    def __init__(self, request, queryset_statistics):

        groups = []
        labels = []
        series = []

        if queryset_statistics.groups:
            for label in queryset_statistics.labels:
                labels.append(label)
            for group in queryset_statistics.groups:
                groups.append(group)
        else:
            for label in queryset_statistics.labels:
                labels.append(label)

        for serie in queryset_statistics.series:
            series.append(serie)

        symbol = queryset_statistics.symbol
        title = queryset_statistics.title

        super(QueryStatisticsChart, self).__init__(request, labels, series, groups, symbol, title)

        self.clickable = self.request is not None
        self.queryset_statistics = queryset_statistics
        self.process_request()

    def process_request(self):
        super(Chart, self).process_request()
        if self.request and 'uuid' in self.request.GET and self.request.GET['uuid'] == self.title:
            note = None
            i, j = self.request.GET['position'].split('x')
            qs = self.queryset_statistics.querysets[int(i)][int(j)]
            title = get_metadata(qs.model, 'verbose_name_plural')
            if qs.count() > 25:
                qs = qs[0:25]
                note = u'Apenas uma amostra com 25 registros está sendo exibida.'
            component = ModelTable(self.request, title, qs, note=note)
            raise ComponentHasResponseException(ComponentResponse(component))


class Timeline(RequestComponent):

    class Media:
        css = {'all': ('/static/css/timeline.css',)}

    def __init__(self, request, description, items=[]):
        super(Timeline, self).__init__(description, request)
        self.items = items
        self.description = description
        self.width = 0

    def add(self, status, date):
        self.items.append((status, date))

    def __str__(self):
        self.width = str(100.0/len(self.items))
        return self.render('timeline.html')


class RoleSelector(RequestComponent):

    def __init__(self, request):
        super(RoleSelector, self).__init__('roleselector', request)
        self.groups = []
        self.process_request()
        self._load()

    def _load(self):
        if self.request.user.is_authenticated:
            from djangoplus.admin.models import Group, Role
            pks = self.request.user.role_set.values_list('group', flat=True).distinct()
            for group in Group.objects.filter(pk__in=pks).distinct():
                group.roles = Role.objects.filter(group=group, user=self.request.user).distinct()
                group.scopes = group.roles.filter(scope__isnull=False).count()
                self.groups.append(group)

    def process_request(self):
        if 'data[]' in self.request.GET:
            pks = []
            for value in self.request.GET.getlist('data[]'):
                if value.strip():
                    pks.append(value)
            self.request.user.role_set.update(active=False)
            self.request.user.role_set.filter(pk__in=pks).update(active=True)
            self.request.user.check_role_groups()
            raise ComponentHasResponseException(HttpResponse())


class QrCode(RequestComponent):
    def __init__(self, request, text, width=200, height=200):
        super(QrCode, self).__init__(text, request)
        self.text = text
        self.width = width
        self.base64 = None

    def __str__(self):
        if self.as_pdf:
            qr = qrcode.QRCode()
            qr.add_data(self.text)
            image = qr.make_image()
            file_path = tempfile.mktemp()
            buffer = open(file_path, 'wb')
            image.save(buffer, format="JPEG")
            self.base64 = base64.b64encode(open(file_path, 'rb').read()).decode('utf-8')
        return super(QrCode, self).__str__()


class ProgressBar(RequestComponent):
    def __init__(self, request, percentual):
        self.percentual = percentual
        super(ProgressBar, self).__init__(percentual, request)


class Table(RequestComponent):
    def __init__(self, request, title, header=list(), rows=list(), footer=list(), enumerable=True, note=None):
        super(Table, self).__init__(title, request)
        self.title = title
        self.header = header
        self.rows = rows
        self.footer = footer
        self.enumerable = enumerable
        self.note = note


class ModelTable(Table):
    def __init__(self, request, title, qs, list_display=(), note=None):

        header = []
        rows = []

        if not list_display:
            list_display = get_metadata(qs.model, 'list_display')

        for lookup in list_display:
            column_name = get_fiendly_name(qs.model, lookup, as_tuple=False)
            header.append(column_name)

        for obj in qs:
            row = []
            for lookup in list_display:
                row.append(getattr2(obj, lookup))
            rows.append(row)

        super(ModelTable, self).__init__(request, title, header, rows, note=note)


class QueryStatisticsTable(Table):
    def __init__(self, request, queryset_statistics):
        if queryset_statistics.groups:
            header = []
            rows = []
            footer = []
            if queryset_statistics.series:
                for i, group in enumerate(queryset_statistics.groups):
                    row = [group]
                    for value in queryset_statistics.series[i]:
                        row.append(value)
                    if len(queryset_statistics.series) > 1:
                        row.append(queryset_statistics.xtotal[i])
                    rows.append(row)
                if len(queryset_statistics.series) > 1:
                    footer = [' '] + queryset_statistics.ytotal + [queryset_statistics.total()]
                    header = [' '] + queryset_statistics.labels + [' ']
                else:
                    header = [' '] + queryset_statistics.labels
        else:
            header = []
            rows = []
            footer = []
            if queryset_statistics.series:
                for i, serie in enumerate(queryset_statistics.series[0]):
                    rows.append([queryset_statistics.labels[i], queryset_statistics.series[0][i]])
                if len(queryset_statistics.series[0]) > 1:
                    footer.append('')
                    footer.append(queryset_statistics.total())
        title = queryset_statistics.title
        super(QueryStatisticsTable, self).__init__(request, title, header, rows, footer, enumerable=False)
        self.queryset_statistics = queryset_statistics
        self.symbol = queryset_statistics.symbol


class ModelReport(RequestComponent):
    def __init__(self, request, title, qs, list_display=(), list_filter=(), distinct=False):
        super(ModelReport, self).__init__(title, request)
        self.title = title
        self.qs = qs
        self.components = []
        self.filters = []

        if list_filter:
            form = forms.Form(request, method='GET')
            form.icon = 'fa-file-text-o'
            form.title = ''
            form.submit_label = 'Gerar Relatório'
            for field_name in list_filter:
                field = get_field(qs.model, field_name)
                if hasattr(field, 'choices') and field.choices:
                    form.fields[field_name] = forms.ChoiceField(choices=[['', '']]+field.choices, label=field.verbose_name, required=False)
                else:
                    form.fields[field_name] = forms.ModelChoiceField(field.remote_field.model.objects.all(), label=field.verbose_name, required=False)
            if form.is_valid():
                for field_name in list_filter:
                    value = form.cleaned_data[field_name]
                    if value:
                        qs = qs.filter(**{field_name : value})
                        self.filters.append((form.fields[field_name].label, value))
            self.form = form
        if distinct:
            pks = qs.values_list('pk', flat=True).order_by('pk').distinct()
            qs = qs.model.objects.filter(pk__in=pks)

        order_by = get_metadata(qs.model, 'order_by', iterable=True)
        if order_by:
            qs = qs.order_by(*order_by)
        table_description = get_metadata(qs.model, 'verbose_name_plural')
        self.table = ModelTable(request, table_description, qs, list_display)

    def count(self, vertical_key, horizontal_key=None, symbol=None, add_table=True, add_chart=False):
        statistics = self.qs.count(vertical_key, horizontal_key)
        statistics.title = 'Resumo'
        statistics.symbol = symbol
        if add_table:
            self.components.append(statistics.as_table(self.request))
        if add_chart:
            self.components.append(statistics.as_chart(self.request))



# -*- coding: utf-8 -*-
from djangoplus.utils import permissions
from djangoplus.ui.components.panel import ModelPanel
from djangoplus.ui.components.paginator import Paginator
from djangoplus.utils.metadata import get_fieldsets, get_metadata
from djangoplus.ui.components.forms import ModelForm, ValidationError


class Relation(object):
    def __init__(self, instance, relation_name):
        self.instance = instance
        self.model = type(self.instance)
        self.relation_name = relation_name
        self.relation_verbose_name = None
        self.relation_model = None
        self.relation_value = None
        self.relation_type = None
        self.hidden_field_name = None

        app_label = get_metadata(self.model, 'app_label')
        model_name = self.model.__name__.lower()

        self.add_label = None
        self.can_add = None
        self.add_url = None
        self.edit_url = None
        self.view_url = None
        self.delete_url = None

        self.is_one_to_one = False
        self.is_many_to_one = False
        self.is_one_to_many = False
        self.is_one_to_many_reverse = False
        self.is_many_to_many = False
        
        attr = getattr(self.model, relation_name)
        descriptor_name = attr.__class__.__name__

        if hasattr(attr, 'related'):
            if descriptor_name == 'ReverseOneToOneDescriptor':
                self.is_one_to_one = True
                self.relation_type = 'OneToOne'
                self.relation_model = attr.related.remote_field.model
                self.relation_verbose_name = getattr(attr.related.target_field.model, '_meta').verbose_name

                self.relation_value = getattr(instance, relation_name)
                relation_app_label = get_metadata(self.model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()

                self.add_url = None
                self.view_url = '/view/{}/{}/{}/'.format(relation_app_label, relation_model_name, self.relation_value.pk)

        elif hasattr(attr, 'field'):

            field_name = attr.field.__class__.__name__

            if descriptor_name == 'ForwardOneToOneDescriptor':
                self.is_one_to_one = True
                self.relation_type = 'OneToOne'
                self.relation_model = attr.field.remote_field.model
                self.relation_verbose_name = attr.field.verbose_name

                self.relation_value = getattr(instance, relation_name)
                relation_app_label = get_metadata(self.relation_model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()

                self.add_url = '/add/{}/{}/{}/{}/'.format(app_label, model_name, self.instance.pk, relation_name)
                if self.relation_value:
                    self.add_url = '/add/{}/{}/{}/{}/{}/'.format(app_label, model_name, self.instance.pk, relation_name, self.relation_value.pk)
                    self.view_url = '/view/{}/{}/{}/'.format(relation_app_label, relation_model_name, self.relation_value.pk)
                    self.delete_url = '/delete/{}/{}/{}/'.format(relation_app_label, relation_model_name, self.relation_value.pk)

            elif descriptor_name == 'ForwardManyToOneDescriptor':
                self.is_many_to_one = True
                self.relation_type = 'ManyToOne'
                self.relation_model = attr.field.remote_field.model
                self.relation_verbose_name = attr.field.verbose_name

                self.relation_value = getattr(instance, relation_name)
                relation_app_label = get_metadata(self.relation_model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()
                if self.relation_value:
                    self.view_url = '/view/{}/{}/{}/'.format(relation_app_label, relation_model_name, self.relation_value.pk)

            elif descriptor_name == 'ManyToManyDescriptor' and field_name == 'OneToManyField':
                self.is_one_to_many = True
                self.relation_type = 'OneToMany'
                self.relation_model = attr.field.remote_field.model
                self.relation_verbose_name = attr.field.verbose_name
                self.add_label = attr.field.add_label

                self.relation_value = getattr(instance, relation_name).all()
                relation_app_label = get_metadata(self.relation_model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()

                self.view_url = '/view/{}/{}/{{}}/'.format(relation_app_label, relation_model_name)
                self.add_url = '/add/{}/{}/{}/{}/'.format(app_label, model_name, self.instance.pk, relation_name)
                self.delete_url = '/delete/{}/{}/{}/{}/{{}}/'.format(app_label, model_name, self.instance.pk, relation_name)

            elif descriptor_name == 'ReverseManyToOneDescriptor':
                self.is_one_to_many_reverse = True
                self.relation_type = 'OneToManyReverse'
                self.relation_model = attr.field.model
                self.relation_verbose_name = getattr(attr.field.model, '_meta').verbose_name_plural

                self.hidden_field_name = attr.rel.field.name
                self.relation_value = getattr(instance, relation_name).all()
                relation_app_label = get_metadata(self.relation_model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()

                self.add_url = '/add/{}/{}/{}/{}/'.format(app_label, model_name, self.instance.pk, relation_name)
                self.edit_url = '/add/{}/{}/{}/{}/{{}}/'.format(app_label, model_name, self.instance.pk, relation_name)
                self.view_url = '/view/{}/{}/{{}}/'.format(relation_app_label, relation_model_name)
                self.delete_url = '/delete/{}/{}/{{}}/'.format(relation_app_label, relation_model_name)

            elif descriptor_name == 'ManyToManyDescriptor':
                self.is_many_to_many = True
                self.relation_type = 'ManyToMany'
                self.relation_model = attr.field.remote_field.model
                self.relation_verbose_name = attr.field.verbose_name
                self.add_label = attr.field.add_label
                self.can_add = attr.field.can_add

                self.relation_value = getattr(instance, relation_name).all()
                relation_app_label = get_metadata(self.relation_model, 'app_label')
                relation_model_name = self.relation_model.__name__.lower()

                self.view_url = '/view/{}/{}/{{}}/'.format(relation_app_label, relation_model_name)
                self.add_url = '/add/{}/{}/{}/{}/'.format(app_label, model_name, self.instance.pk, relation_name)
                self.delete_url = '/delete/{}/{}/{}/{}/{{}}/'.format(app_label, model_name, self.instance.pk, relation_name)
            else:
                raise Exception()
        elif hasattr(attr, '_metadata'):
            self.relation_value = getattr(self.instance, self.relation_name)()
            if hasattr(self.relation_value, 'model'):
                self.relation_model = self.relation_value.model
            else:
                self.relation_model = type(self.relation_value)
            self.relation_verbose_name = getattr(attr, '_metadata')['{}:verbose_name'.format(self.relation_name)]

            relation_app_label = get_metadata(self.relation_model, 'app_label')
            relation_model_name = self.relation_model.__name__.lower()
            self.view_url = '/view/{}/{}/{{}}/'.format(relation_app_label, relation_model_name)
        else:
            raise Exception()
        # self.debug()

    def get_component(self, request, as_pdf=False):
        verbose_name = getattr(self.relation_model, '_meta').verbose_name
        if self.is_one_to_one or self.is_many_to_one:
            panel_fieldsets = getattr(self.relation_model, 'fieldsets', None)
            if panel_fieldsets:
                panel_fieldsets = ((self.relation_verbose_name, panel_fieldsets[0][1]),)
            else:
                panel_fieldsets = get_fieldsets(self.relation_model, self.relation_verbose_name)
            component = ModelPanel(request, self.relation_value or self.relation_model(), fieldsets=panel_fieldsets, complete=False)
            if self.view_url and permissions.has_view_permission(request, self.relation_model):
                label = 'Detalhar {}'.format(verbose_name)
                component.drop_down.add_action(label, self.view_url, 'ajax', 'fa-eye', category=label)
            if self.add_url and permissions.has_edit_permission(request, self.model):
                label = 'Atualizar {}'.format(verbose_name)
                component.drop_down.add_action(label, self.add_url, 'popup', 'fa-edit', category=label)
            if self.delete_url and permissions.has_edit_permission(request, self.model):
                label = 'Excluir {}'.format(verbose_name)
                component.drop_down.add_action(label, self.delete_url, 'popup', 'fa-close', category=label)
        else:
            inlines = []
            fieldsets = getattr(self.model, 'fieldsets', ())
            title = self.relation_verbose_name
            for fieldset in fieldsets:
                fieldset_relations = fieldset[1].get('relations', ())
                fieldset_inlines = fieldset[1].get('inlines', ())
                fieldset_fields = fieldset[1].get('fields', ())
                for inline in fieldset_inlines:
                    inlines.append(inline)
                if self.relation_name in fieldset_relations or self.relation_name in fieldset_inlines or self.relation_name in fieldset_fields:
                    if len(fieldset_relations) + len(fieldset_inlines) + len(fieldset_fields) == 1:
                        title = fieldset[0].split('::')[-1]

            if self.is_one_to_many or self.is_many_to_many:
                if self.can_add:
                    has_add_permission = permissions.check_group_or_permission(request, self.can_add)
                else:
                    has_add_permission = permissions.has_add_permission(request, self.model)
            else:
                has_add_permission = permissions.has_add_permission(request, self.relation_model)

            component = Paginator(request, self.relation_value.all(request.user), title, relation=self, list_subsets=[], readonly=not has_add_permission)
            component.add_actions()
            if self.add_url and has_add_permission:
                if self.relation_name in inlines:
                    instance = self.relation_model()
                    setattr(instance, self.hidden_field_name, self.instance)
                    if not hasattr(instance, 'can_add') or instance.can_add():
                        form_name = get_metadata(self.relation_model, 'add_form')
                        if form_name:
                            fromlist = get_metadata(self.relation_model, 'app_label')
                            forms_module = __import__('{}.forms'.format(fromlist), fromlist=list(map(str, [fromlist])))
                            Form = getattr(forms_module, form_name)
                        else:
                            class Form(ModelForm):
                                class Meta:
                                    model = self.relation_model
                                    fields = get_metadata(self.relation_model, 'form_fields', '__all__')
                                    exclude = get_metadata(self.relation_model, 'exclude_fields', ())
                                    submit_label = 'Adicionar'
                                    title = 'Adicionar {}'.format(get_metadata(self.relation_model, 'verbose_name'))
                        form = Form(request, instance=instance, inline=True)
                        if self.hidden_field_name in form.fields:
                            del (form.fields[self.hidden_field_name])
                        component.form = form
                        if form.is_valid():
                            try:
                                form.save()
                                component.message = 'Ação realizada com sucesso'
                            except ValidationError as e:
                                form.add_error(None, str(e.message))
                else:
                    add_label = self.add_label or get_metadata(self.relation_model, 'add_label')
                    label = add_label or 'Adicionar {}'.format(verbose_name)
                    component.add_action(label, self.add_url, 'popup', 'fa-plus')

        component.as_pdf = as_pdf

        return component

    def debug(self):
        url = 'http://localhost:8000'
        print('Instance:', self.instance)
        print('Model:', self.model)
        print('Relation Name:', self.relation_name)
        print('Relation Verbose Name:', self.relation_verbose_name)
        print('Relation Model:', self.relation_model)
        print('Relation Value:', self.relation_value)
        print('Relation Type:', self.relation_type)
        if self.add_url:
            print('Add URL', '{}{}'.format(url, self.add_url))
        if self.edit_url:
            print('Edit URL', '{}{}'.format(url, self.edit_url))
        if self.view_url:
            print('View URL', '{}{}'.format(url, self.view_url))
        if self.delete_url:
            print('Delete URL', '{}{}'.format(url, self.delete_url))
        print('\n\n\n\n')


# -*- coding: utf-8 -*-
import os
import time
import datetime
import traceback
from selenium import webdriver
from selenium.webdriver.firefox.options import Options
from selenium.common.exceptions import WebDriverException


class Browser(webdriver.Firefox):
    def __init__(self, server_url, options=None, verbose=True, slowly=False, maximize=True):

        if not options:
            options = Options()
        if maximize:
            options.add_argument("--start-maximized")
        else:
            options.add_argument("--window-size=720x800")
        if 'HEADLESS' in os.environ:
            options.add_argument("--headless")

        super(Browser, self).__init__(options=options)

        self.verbose = verbose
        self.slowly = slowly
        self.watched = False
        self.server_url = server_url

        if maximize:
            self.maximize_window()
        else:
            self.set_window_position(700, 0)
            self.set_window_size(720, 800)
        self.switch_to.window(self.current_window_handle)

    def wait(self, seconds=1):
        seconds = self.slowly and 3 or seconds
        time.sleep(seconds)

    def watch(self, e):
        if self.watched:
            raise e
        else:
            traceback.print_exc()
            self.watched = True
            self.save_screenshot('/tmp/test.png')
            if 'HEADLESS' not in os.environ:
                input('Type enter to continue...')

    def print(self, message):
        if self.verbose:
            print(message)

    def open(self, url):
        self.get("{}{}".format(self.server_url, url))

    def enter(self, name, value, submit=False, count=2):

        if callable(value):
            value = value()
        if type(value) == datetime.date:
            value = value.strftime('%d/%m/%Y')
        if value:
            self.print('{} "{}" for "{}"'.format('Entering', value, name))
        try:
            if submit:
                self.execute_script("enter('{}', '{}', 1)".format(name, value))
            else:
                self.execute_script("enter('{}', '{}')".format(name, value))
        except WebDriverException as e:
            if count:
                self.wait()
                self.enter(name, value, submit, count-1)
            else:
                self.watch(e)
                
    def choose(self, name, value, count=2):
        self.print('{} "{}" for "{}"'.format('Choosing', value, name))
        try:
            headless = 'HEADLESS' in os.environ and 'true' or 'false'
            self.execute_script("return choose('{}', '{}', {})".format(name, value, headless))
            self.wait(2)
        except WebDriverException as e:
            if count:
                self.wait()
                self.choose(name, value, count-1)
            else:
                self.watch(e)

    def dont_see_error_message(self, testcase=None):
        elements = self.find_elements_by_class_name('alert-danger')
        if elements:
            messages = [element.text for element in elements]
            if 'HEADLESS' not in os.environ:
                input('Type enter to continue...')
            elif testcase:
                exception_message = 'The following messages were found on the page: {}'.format(';'.join(messages))
                raise testcase.failureException(exception_message)

    def look_for(self, text, count=2):
        self.print('Looking for "{}"'.format(text))
        try:
            assert text in self.find_element_by_tag_name('body').text
        except WebDriverException as e:
            if count:
                self.wait()
                self.look_for(text, count-1)
            else:
                self.watch(e)

    def look_at_popup_window(self, count=2):
        self.print('Looking at popup window')
        try:
            self.execute_script("lookAtPopupWindow()")
        except WebDriverException as e:
            if count:
                self.wait()
                self.look_at_popup_window(count-1)
            else:
                self.watch(e)

    def look_at(self, text, count=2):
        self.print('Loking at "{}"'.format(text))
        try:
            self.execute_script("lookAt('{}')".format(text))
        except WebDriverException as e:
            if count:
                self.wait()
                self.look_at(text, count-1)
            else:
                self.watch(e)

    def look_at_panel(self, text, count=2):
        self.print('Looking at panel "{}"'.format(text))
        try:
            self.execute_script("lookAtPanel('{}')".format(text))
        except WebDriverException as e:
            if count:
                self.wait()
                self.look_at_panel(text, count-1)
            else:
                self.watch(e)

    def click_menu(self, *texts):
        self.print('Clicking menu "{}"'.format('->'.join(texts)))
        self.wait()
        for text in texts:
            self.wait()
            try:
                self.execute_script("clickMenu('{}')".format(text.strip()))
            except WebDriverException as e:
                self.watch(e)
        self.wait()

    def click_link(self, text):
        self.print('Clicking link "{}"'.format(text))
        try:
            self.execute_script("clickLink('{}')".format(text))
        except WebDriverException as e:
            self.watch(e)
        self.wait()

    def click_button(self, text):
        self.print('Clicking button "{}"'.format(text))
        try:
            self.execute_script("clickButton('{}')".format(text))
        except WebDriverException as e:
            self.watch(e)
        self.wait()
        self.dont_see_error_message()
        
    def click_tab(self, text):
        self.print('Clicking tab "{}"'.format(text))
        try:
            self.execute_script("clickTab('{}')".format(text))
        except WebDriverException as e:
            self.watch(e)
        self.wait()

    def click_icon(self, name):
        self.print('Clicking icon "{}"'.format(name))
        try:
            self.execute_script("clickIcon('{}')".format(name))
        except WebDriverException as e:
            self.watch(e)
        self.wait()

    def logout(self):
        self.print('Logging out')
        self.click_icon('Configurações')
        self.wait()
        self.click_link('Sair')
        self.wait()

    def close(self, seconds=0):
        self.wait(seconds)
        super(Browser, self).close()




# -*- coding: utf-8 -*-
import os
import json
from djangoplus.admin.models import Settings
import urllib
from djangoplus.cache import loader
from django.contrib import auth
from djangoplus.utils.aescipher import decrypt
from django.http.response import HttpResponse
from djangoplus.decorators.views import view, action
from djangoplus.ui.components.navigation.breadcrumbs import httprr
from djangoplus.admin.models import User, Unit, Organization
from djangoplus.ui.components.panel import DashboardPanel
from djangoplus.admin.forms import ProfileForm, ChangePasswordForm, SettingsForm, LoginForm, RecoverPassowordForm


@view('Public', login_required=False)
def public(request):
    app_settings = Settings.default()
    if not app_settings.background:
        return httprr(request, '/admin/')
    return locals()


@view('Principal', login_required=True)
def index(request):
    widget_panel = DashboardPanel(request)
    return locals()


@view('Acesso ao Sistema', login_required=False, template='login/login.html')
def login(request, scope=None, organization=None, unit=None):
    auth.logout(request)
    can_register = loader.signup_model is not None and 'HEADLESS' not in os.environ
    organization = organization and Organization.objects.get(pk=organization) or None
    unit = unit and Unit.objects.get(pk=unit) or None
    form = LoginForm(request, scope, organization, unit)
    if form.is_valid():
        return form.submit()
    return locals()


@view('Logout', login_required=False)
def logout(request):
    url = '/'
    unit_id = request.session.get('unit_id', None)
    if unit_id:
        url = '/admin/login/{}/'.format(unit_id)
    auth.logout(request)
    return httprr(request, url, 'Logout realizado com sucesso.')


@view('404')
def error(request):
    return locals()


@view('Alteração de Senha', login_required=False, template='login/password.html')
def password(request, pk, token):
    title = 'Alterar Senha'
    user = User.objects.get(pk=pk, password=decrypt(token))
    form = ChangePasswordForm(request, instance=user)
    if form.is_valid():
        form.save()
        return httprr(request, '/admin/', 'Parabéns! Sua senha foi alterada com sucesso.')
    return locals()


@view('Recuperação de Senha', login_required=False, template='login/recover.html')
def recover(request):
    title = 'Recuperar Senha'
    form = RecoverPassowordForm(request)
    if form.is_valid():
        return httprr(request, '/admin/login/', 'O link para redefinição da senha foi enviado para o e-mail informado.')
    return locals()


@view('Register', login_required=False)
def register(request, token=None, userid=None):
    from djangoplus.ui.components import forms
    from djangoplus.utils.metadata import get_metadata

    initial = {}
    username_field = get_metadata(loader.signup_model, 'role_username')
    email_field = get_metadata(loader.signup_model, 'role_email')
    name_field = get_metadata(loader.signup_model, 'role_name')

    if not loader.signup_model:
        return httprr(request, '/admin/login/', 'O cadastrado externo não está habilitado.')

    if token:
        if token and userid:
            url = 'https://graph.facebook.com/{}?fields=email,first_name,last_name&access_token={}'.format(userid, token)
        elif token:
            url = 'https://www.googleapis.com/oauth2/v1/userinfo?alt=json&access_token={}'.format(token)
        data = json.loads(urllib.request.urlopen(url).read())
        qs = User.objects.filter(username=data['email'])
        if qs.exists():
            user = qs[0]
            auth.login(request, user)
            return httprr(request, '/admin/', 'Usuário autenticado com sucesso.')
        else:
            initial = {name_field: data['name'], username_field : data['email'], email_field : data['email']}

    class RegisterForm(forms.ModelForm):
        class Meta:
            model = loader.signup_model
            fields = get_metadata(loader.signup_model, 'form_fields', '__all__')
            exclude = get_metadata(loader.signup_model, 'exclude_fields', ())
            submit_label = 'Cadastrar'
            title = 'Cadastro de {}'.format(get_metadata(loader.signup_model, 'verbose_name'))
            icon = get_metadata(loader.signup_model, 'icon', None)


    form = RegisterForm(request, initial=initial)
    form.fields[username_field].help_text='Utilizado para acessar o sistema.'

    save_instance = True
    for field_name in form.fields:
        if not initial.get(field_name):
            save_instance = False
    if save_instance:
        instance = loader.signup_model()
        for field_name in form.fields:
            setattr(instance, field_name, initial[field_name])
        instance.save()
        user = User.objects.get(username=initial[username_field])
        auth.login(request, user)
        return httprr(request, '/admin/', 'Usuário cadastrado com sucesso.')

    if form.is_valid():
        instance = form.save()
        extra = email_field and 'Um e-mail será enviado para você tão logo sua conta seja ativada.' or ''
        if instance:
            user = User.objects.get(username=form.cleaned_data[username_field])
            auth.login(request, user)
            return httprr(request, '/admin/', 'Usuário cadastrado com sucesso.')
        else:
            return httprr(request, '..', 'Acesse o link enviado para seu e-mail para confirmar a criação da sua conta.')
    return locals()


@view('Profile')
def profile(request):
    form = ProfileForm(request, instance=request.user)
    if form.is_valid():
        form.save()
        return httprr(request, '..', 'Perfil atualizado com sucesso')
    return locals()


@view('Configure')
def configure(request):
    if not request.user.is_superuser:
        return httprr(request, '/', 'Você não tem permissão para realizar isto!', 'error')
    title = 'Configurações'
    form = SettingsForm(request)
    if form.is_valid():
        form.save()
        return httprr(request, '..', 'Configuração salva com sucesso')
    return locals()


@view('Sidebar')
def toggle_menu(request):
    if 'hidden_menu' in request.session and request.session['hidden_menu']:
        request.session['hidden_menu'] = False
    else:
        request.session['hidden_menu'] = True
    request.session.save()
    return HttpResponse()


@action(User, 'Login as User', can_execute='Superuser')
def login_as(request, pk):
    auth.logout(request)
    user = User.objects.get(pk=pk)
    auth.login(request, user)
    return httprr(request, '/admin/', 'Login realizado com sucesso')







# Generated by Django 2.0.3 on 2018-04-16 10:34

from django.db import migrations, models
import djangoplus.db.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0001_squashed_0015_settings_company'),
    ]

    operations = [
        migrations.CreateModel(
            name='Scope',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ascii', djangoplus.db.models.fields.SearchField(blank=True, default='', editable=False)),
            ],
            options={
                'verbose_name': 'Scope',
                'verbose_name_plural': 'Scopes',
            },
        ),
    ]



# -*- coding: utf-8 -*-
import re
import platform
import random
import curses
import time  #: 3
import os
import signal
from time import sleep
from subprocess import Popen, PIPE, DEVNULL
from djangoplus.tutorial.utils import display_subtitle, simulate_type, simulate_command_type


class Subtitle(object):

    def display(self, message, duration=4):
        display_subtitle(message, duration)


class Terminal(object):

    def __init__(self, verbose=True, python='python'):
        self.proccess = None
        self.verbose = verbose
        self.python = python

    def execute(self, command, clear=True, base_dir=None):
        if clear:
            os.system('clear')
            simulate_type('', shell=True)
        if self.verbose:
            simulate_command_type(command, shell=True)
        if command.startswith('python'):
            command.replace('python', self.python)
        if base_dir:
            command = 'cd {} && {}'.format(base_dir, command)
        if not self.verbose:
            command = '{}  > /dev/null'.format(command)
        os.system(command)

    def show(self, visible=True):
        minimize_terminal_script = '''
            tell application "Terminal"
              set miniaturized of window 1 to {}
            end tell
        '''.format(visible and 'false' or 'true')
        self.proccess = Popen(['osascript', '-'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        self.proccess.communicate(minimize_terminal_script.encode())

    def hide(self):
        self.show(False)


class HttpServer(object):
    def __init__(self, base_dir, verbose=True, python='python'):
        self.process = None
        self.base_dir = base_dir
        self.verbose = verbose
        self.python = python

    def start(self):
        command = 'cd {} && {} manage.py runserver'.format(self.base_dir, self.python)
        if self.verbose:
            os.system('clear')
            simulate_type('python manage.py runserver', shell=True)
            stdout = PIPE
        else:
            stdout = DEVNULL
        self.process = Popen(command, stdout=stdout, stderr=PIPE, shell=True, preexec_fn=os.setsid)
        sleep(5)

    def stop(self):
        if self.verbose:
            print('^C')
        if self.process:
            os.killpg(os.getpgid(self.process.pid), signal.SIGTERM)


# apt-get install ffmpeg
class VideoRecorder(object):
    def __init__(self):
        self.proccess = None

    def start(self):
        if 'darwin' in platform.system().lower():
            list_divices_procces = Popen('ffmpeg -f avfoundation -list_devices true -i ""'.split(), stdout=PIPE, stderr=PIPE)
            output, err = list_divices_procces.communicate()
            i = 0
            for line in err.decode('utf-8').split('\n'):
                if 'capture screen' in line.lower():
                    for i in range(0, 3):
                        if '[{}]'.format(i) in line:
                            break
            self.proccess = Popen(['ffmpeg', '-y', '-f', 'avfoundation', '-i', str(i), '/tmp/video.mkv'], stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL)
        else:
            xrandr = Popen("xrandr".split(), stdout=PIPE)
            sizes = xrandr.stdout.readlines()
            size = re.findall(r'\d+x\d+', [line.decode('utf-8').strip().split()[0] for line in sizes if '*' in line.decode('utf-8')][0])[0]
            cmd = 'ffmpeg -video_size {} -framerate 25 -f x11grab -i :0.0+0,0 -f pulse -ac 2 -i default /tmp/video.mkv'
            self.proccess = Popen(cmd.format(size).split(), stdin=DEVNULL, stdout=DEVNULL, stderr=DEVNULL)
        sleep(5)

    def stop(self, file_name=None, audio_file_path=None):
        os.kill(self.proccess.pid, signal.SIGTERM)
        sleep(2)
        tmp_file_path = '/tmp/video.mkv'
        desktop_path = os.path.join(os.path.join(os.path.expanduser('~')), 'Desktop')
        if not os.path.exists(desktop_path):
            desktop_path = os.path.join(os.path.expanduser('~'))
        output_file_path = '{}/{}.mkv'.format(desktop_path, file_name)
        if file_name:
            if os.path.exists(output_file_path):
                os.unlink(output_file_path)
            os.rename(tmp_file_path, output_file_path)
            print(output_file_path)
            if audio_file_path and 'darwin' in platform.system().lower():
                output_audio_path = '{}/tmp{}.mkv'.format(desktop_path, file_name)
                combine_audio_cmd = 'ffmpeg -y -i {} -i {} -c copy -map 0:0 -map 1:0 -shortest {}  > /dev/null 2>&1'
                os.system(combine_audio_cmd.format(output_file_path, audio_file_path, output_audio_path))
                sleep(5)
                os.rename(output_audio_path, output_file_path)


class EditorSimulator(object):

    COMMENT_MARKUP = '{}{} '.format('#', ':')

    def __init__(self, file_path=None, slow=True):  #: 0 Definindo o construtor
        self.screen = None  #: 0
        self.height = 0  #: 0
        self.width = 0  #: 0
        self.top = []  #: 0
        self.center = []  #: 0
        self.bottom = []  #: 0
        self.steps = {}  #: 0
        self.slow = slow  #: 0

        if file_path:  #: 0
            self.lines = open(file_path).readlines()  #: 0
        else:  #: 0
            self.lines = []  #: 0

        self.initialize_curses()

    def initialize_curses(self):
        self.screen = curses.initscr()
        self.height, self.width = self.screen.getmaxyx()

        curses.start_color()
        curses.use_default_colors()
        curses.init_pair(1, curses.COLOR_GREEN, -1)
        curses.curs_set(0)

        self.top = []
        self.center = []
        self.bottom = []
        self.steps = {}

    def proccess_lines(self, start_step=None, end_step=None):
        i = 0
        for j, line in enumerate(self.lines):
            tokens = line.split(EditorSimulator.COMMENT_MARKUP)
            if len(tokens) > 1:
                step_number = float(tokens[1].strip().split()[0])
                if j and not self.lines[j - 1].strip():
                    self.lines[j - 1] = '{}{}{}'.format(self.lines[j - 1], EditorSimulator.COMMENT_MARKUP, step_number)
        for j, text in enumerate(self.lines):
            j += 1
            tokens = text.split(EditorSimulator.COMMENT_MARKUP)
            text = tokens[0].rstrip()
            if len(tokens) > 1:
                step = tokens[1].strip()
                step_number = float(step.split()[0])
                step_message = ' '.join(step.split()[1:]).strip()
                if start_step and step_number < start_step:
                    step_number = None
                    step_message = ''
            else:
                step_number = None
                step_message = ''
            if step_number is None:
                if i < self.height-1:
                    self.center.append(text)
                    self.screen.addstr(i, 0, text[0:self.width-1])
                    i += 1
                else:
                    self.bottom.append(text)
            else:
                if step_number not in self.steps:
                    self.steps[step_number] = []
                for line in self.lines[:j]:
                    tokens = line.replace('\n', '').split(EditorSimulator.COMMENT_MARKUP)
                    if len(tokens) > 1 and float(tokens[1].strip().split()[0]) > step_number:
                        j -= 1
                self.steps[step_number].append([text, j, step_message])

        self.screen.refresh()

    def simulate(self, start_step=None, end_step=None, pause=False):  #: 1 Definindo o método "simulate"
        if not self.screen:  #: 1
            self.initialize_curses()  #: 1
        self.proccess_lines(start_step, end_step)  #: 1
        for step in sorted(self.steps.keys()):  #: 1
            if end_step is not None and step > float(end_step):  #: 1
                break  #: 1
            for text, i, message in self.steps[step]:  #: 1
                if message:  #: 1
                    display_subtitle(message, self.slow and 3 or 1)  #: 1
                self.write(text, i)  #: 1
        self.close(validate=end_step is None, pause=pause)  #: 1

    def scroll(self, i):
        time.sleep(random.random() * 10.0 / (self.slow and 400 or 50))
        first_visible_line = len(self.top)
        # text goes up
        if i > first_visible_line:
            for j in range(first_visible_line, i):
                text = self.center.pop(0)
                self.top.append(text)
                self.screen.move(0, 0)
                self.screen.deleteln()
                if self.bottom:
                    text = self.bottom.pop(0)
                    self.center.append(text)
                    self.screen.addstr(len(self.center)-1, 0, text)
                self.screen.refresh()
                time.sleep(0.1)
        # text goes down
        elif i < first_visible_line:
            self.screen.move(0, 0)
            for j in range(i, first_visible_line):
                if self.top:
                    text = self.top.pop()
                    self.center.insert(0, text)
                    self.screen.insertln()
                    self.screen.addstr(0, 0, text)
                    if len(self.center) > self.height-1:
                        text = self.center.pop()
                        self.bottom.insert(0, text)
                    self.screen.refresh()
                    time.sleep(0.1)
        self.screen.refresh()

    def write(self, text, i, start=0):
        middle = int(self.height/2)
        if len(self.center) > middle and i > middle:
            shift = middle
            self.scroll(i-shift-1)
        else:
            shift = i - 1
            self.scroll(i - shift - 1)
        self.screen.move(0 + shift, 0)
        curses.setsyx(0 + shift, 0)
        if start:
            self.center[0] = '{}{}{}'.format(self.center[0][:start], text, self.center[0][start:])
        else:
            self.screen.insdelln(1)
            self.center.insert(0+shift, text)
            if len(self.center) == self.height:
                self.bottom.insert(0, self.center.pop())
        for j, c in enumerate(text):
            if j > 0:
                curses.curs_set(1)
            if start:
                self.screen.insstr(0+shift, start+j, c, curses.color_pair(1))
            else:
                self.screen.addstr(0+shift, j, c, curses.color_pair(1))
            time.sleep(random.random() * 10.0 / (self.slow and 50 or 400))
            self.screen.refresh()
            curses.curs_set(0)

    def save(self, file_path):  #: 4 Definindo o método "save"
        tmp = open(file_path, 'w')  #: 4
        tmp.write(self.get_content())  #: 4
        tmp.flush()  #: 4
        tmp.close()  #: 4

    def get_content(self):
        l = list()
        if self.top:
            l.append('\n'.join(self.top))
            l.append('\n')
        l.append('\n'.join(self.center))
        if self.bottom:
            l.append('\n')
            l.append('\n'.join(self.bottom))
        return ''.join(l)

    def validate(self):
        l = list()
        for line in self.lines:
            l.append(line.split(EditorSimulator.COMMENT_MARKUP)[0].rstrip())
        original_content = '\n'.join(l)
        content = self.get_content()
        if not original_content == content:
            open('/tmp/output0.txt', 'w').write(original_content)
            open('/tmp/output1.txt', 'w').write(content)
            raise Exception('Output content does not mach original content!!!')

    def close(self, pause=False, validate=False):  #: 2 Definindo a função "close"
        if pause:  #: 2
            self.screen.getch()  #: 2
        if validate:  #: 2
            self.validate()  #: 2
        if self.screen:
            self.screen.clear()  #: 2
            curses.endwin()  #: 2
            sleep(1)  #: 2
        self.screen = None  #: 2



# -*- coding: utf-8 -*-

# DJANGO FIELDS
import sys
from django.utils.text import camel_case_to_spaces
from django.utils.safestring import mark_safe


def is_many_to_many(model, name):
    attr = hasattr(model, name) and getattr(model, name) or None
    return attr and hasattr(attr, 'field') and attr.field.__class__.__name__ == 'ManyToManyField' or False


def is_one_to_many(model, name):
    attr = hasattr(model, name) and getattr(model, name) or None
    return attr and hasattr(attr, 'field') and attr.field.__class__.__name__ == 'OneToManyField' or False


def is_one_to_one(model, name):
    attr = hasattr(model, name) and getattr(model, name) or None
    return attr and hasattr(attr, 'field') and attr.field.__class__.__name__ == 'OneToOneField' or False


def is_many_to_one(model, name):
    attr = hasattr(model, name) and getattr(model, name) or None
    return attr and attr.__class__.__name__ == 'ForwardManyToOneDescriptor' or False


def is_one_to_many_reverse(model, name):
    attr = hasattr(model, '{}_set'.format(name)) and getattr(model, '{}_set'.format(name)) or None
    return attr and attr.__class__.__name__ == 'ManyToManyDescriptor' or False


def list_m2m_fields_with_model(cls):
    return [(f, f.model if f.model != cls else None) for f in cls._meta.get_fields() if f.many_to_many and not f.auto_created]


def list_related_objects(cls):
    return [f for f in cls._meta.get_fields() if (f.one_to_many or f.one_to_one) and f.auto_created]


def find_field_by_name(model, field_name):
    return model._meta.get_field(field_name)


def list_field_names(model, include_m2m=False):
    l = [field.name for field in model._meta.local_fields]
    if include_m2m:
        for field in model._meta.local_many_to_many:
            l.append(field.name)
    return l


def get_field(model, lookup):
    tokens = lookup.split('__')
    tokens.reverse()
    field_name = tokens.pop()
    field = find_field_by_name(model, field_name)
    return get_field_recursively(field, tokens)


def get_field_recursively(field, tokens):
    if tokens:
        field_name = tokens.pop()
        if field.remote_field and field.remote_field.model:
            field = find_field_by_name(field.remote_field.model, field_name)
        else:
            field = find_field_by_name(field.related_model, field_name)
        return get_field_recursively(field, tokens)
    else:
        return field


def get_fieldsets(model, title='Dados Gerais'):
    fields = []
    for field in get_metadata(model, 'fields')[1:]:
        if not field.name.endswith('_ptr') and not field.name == 'ascii' and not field.name == 'tree_index':
            fields.append(field.name)

    for field in get_metadata(model, 'local_many_to_many'):
        fields.append(field.name)

    fieldsets = ((title, dict(fields=fields)),)
    return fieldsets


def get_fiendly_name(model_or_field, lookup, as_tuple=False):
    l = []
    tokens = lookup.split('__')
    tokens.reverse()
    pronoun = None
    to = None
    while (tokens):
        token = tokens.pop()

        if hasattr(model_or_field, token):
            field_or_function = getattr(model_or_field, token)
            if hasattr(field_or_function, 'field'):
                field_or_function = field_or_function.field
            elif hasattr(field_or_function, 'field_name'):
                field_or_function = model_or_field._meta.get_field(token)
        else:
            field_or_function = model_or_field

        if pronoun and l:
            l.append(pronoun)

        if hasattr(field_or_function, 'remote_field') and field_or_function.remote_field and field_or_function.remote_field.model:
            pronoun = get_metadata(field_or_function.remote_field.model, 'verbose_female') and 'da' or 'do'

            if tokens:
                model_or_field = field_or_function.remote_field.model
                name = field_or_function.verbose_name
            else:
                model_or_field = field_or_function
                name = model_or_field.verbose_name
            sortable = True
            to = field_or_function.remote_field.model

        elif hasattr(field_or_function, 'verbose_name'):
            model_or_field = field_or_function
            name = model_or_field.verbose_name
            sortable = True

        elif hasattr(field_or_function, '_metadata'):
            model_or_field = field_or_function
            name = field_or_function._metadata['{}:verbose_name'.format(token)]
            sortable = False
        else:
            model_or_field = field_or_function
            name = token
            sortable = False
        l.append(str(name))
    l.reverse()
    verbose_name = ' '.join(l)
    verbose_name = l[0]
    return as_tuple and (verbose_name, lookup, sortable, to) or verbose_name

# META DATA


def set_metadata(cls, attr, value):
    if not hasattr(cls, '_metadata'):
        cls._metadata = dict()
    cls._metadata['{}:{}'.format(cls.__name__, attr)] = value


def get_metadata(model, attr, default=None, iterable=False):
    _metadata = None

    if not hasattr(model._meta, attr):
        field_attr = None

        if attr == 'search_fields':
            field_attr = 'search'
        elif attr == 'exclude_fields':
            field_attr = 'exclude'
        elif attr == 'list_filter':
            field_attr = 'filter'

        if field_attr:
            _metadata = []
            fields = []
            for field in model._meta.fields:
                fields.append(field)
            for m2m in list_m2m_fields_with_model(model):
                fields.append(m2m[0])
            for field in fields:
                if hasattr(field, field_attr):
                    value = getattr(field, field_attr)
                    if value:
                        if type(value) == bool and value:
                            _metadata.append(field.name)
                        elif type(value) in (list, tuple) and value:
                            for lookup in value:
                                if lookup in ('pk', 'id'):
                                    _metadata.append(field.name)
                                else:
                                    _metadata.append('{}__{}'.format(field.name, lookup))
            setattr(model._meta, field_attr, _metadata)
    else:
        _metadata = getattr(model._meta, attr)
        if callable(_metadata):
            _metadata = _metadata()

    if _metadata is None:
        if attr == 'list_display':
            _metadata = []
            fields = []
            for field in model._meta.fields:
                fields.append(field)
            for field in model._meta.local_many_to_many:
                fields.append(field)
            for field in fields[1:6]:
                if not field.name.endswith('_ptr') and not field.name == 'ascii' and not field.name == 'tree_index':
                    _metadata.append(field.name)
        else:
            _metadata = default

    if attr == 'fieldsets' and hasattr(model, 'fieldsets'):
        _metadata = model.fieldsets or default
    elif attr == 'view_fieldsets':
        if hasattr(model, 'view_fieldsets'):
            _metadata = model.view_fieldsets
        else:
            _metadata = hasattr(model, 'fieldsets') and model.fieldsets or default

    check_recursively = not _metadata
    if attr in ('verbose_name', 'verbose_name_plural'):
        if str(_metadata) in (camel_case_to_spaces(model.__name__), '{}s'.format(camel_case_to_spaces(model.__name__))):
            check_recursively = True
    if attr in ('list_menu', 'verbose_female', 'class_diagram'):
        check_recursively = False

    if check_recursively:
        if hasattr(model, '__bases__') and model.__bases__ and hasattr(model.__bases__[0], '_meta'):
            _metadata = get_metadata(model.__bases__[0], attr)

    if iterable:
        if _metadata:
            if not type(_metadata) in (tuple, list):
                _metadata = _metadata,
        else:
            _metadata = ()

    return _metadata

# ROLE


def get_scope(model, organization_model, unit_model):
    role_scope = get_metadata(model, 'role_scope')
    if role_scope:
        attr = getattr2(model, role_scope)
        if hasattr(attr, 'model'):
            return get_metadata(attr.model, 'verbose_name')
        else:
            field = get_field(model, role_scope)
            return get_metadata(field.remote_field.model, 'verbose_name')
    return None


def check_role(self, saving=True):
    role_name = get_metadata(self.__class__, 'role_name')
    role_username = get_metadata(self.__class__, 'role_username')
    verbose_name = get_metadata(self.__class__, 'verbose_name')
    role_email = get_metadata(self.__class__, 'role_email', '')
    role_scope = get_metadata(self.__class__, 'role_scope', False)
    role_notify = get_metadata(self.__class__, 'role_notify', False)

    if role_username:
        from django.conf import settings
        from django.contrib.auth.models import Group
        from djangoplus.admin.models import User, Role, Organization, Unit

        group_name = verbose_name
        username = getattr2(self, role_username)
        name = role_name and getattr2(self, role_name) or None
        email = role_email and getattr2(self, role_email) or ''

        scopes = []

        if username:
            if role_scope:
                value = getattr2(self, role_scope)
                if hasattr(value, 'all'):
                    for scope in value.all():
                        scopes.append(scope)
                elif value:
                    scopes.append(value)

            group = Group.objects.get_or_create(name=group_name)[0]

            if saving:
                qs = User.objects.filter(username=username)
                if qs.exists():
                    user = qs[0]
                    user.email = email
                    user.name = name or str(self)
                    user.save()
                else:
                    user = User()
                    user.username = username
                    user.name = name or str(self)
                    user.email = email
                    user.save()
                    if user.email and not (settings.DEBUG or 'test' in sys.argv):
                        user.send_access_invitation()

                already_in_group = user.groups.filter(pk=group.pk).exists()
                if not already_in_group:
                    user.groups.add(group)
                if scopes:
                    for scope in scopes:
                        Role.objects.get_or_create(user=user, group=group, scope=scope)
                else:
                    Role.objects.get_or_create(user=user, group=group)
                if role_notify and not already_in_group:
                    user.send_access_invitation_for_group(group)
            else:
                user = User.objects.filter(username=username).first()
                if user:
                    if scopes:
                        for scope in scopes:
                            Role.objects.filter(user=user, group=group, scope=scope).delete()
                    else:
                        Role.objects.filter(user=user, group=group).delete()
                    user.check_role_groups()


# REFLECTION


def iterable(string_or_iterable):
    if string_or_iterable and not type(string_or_iterable) in (list, tuple):
        return string_or_iterable,
    return string_or_iterable


def getattr2(obj, args):
    if args == '__str__':
        splitargs = [args]
    else:
        splitargs = args.split('__')
    return getattr_rec(obj, splitargs)


def getattr_rec(obj, args):
    if not obj:
        return None
    if callable(obj):
        obj = obj()
    if len(args) > 1:
        attr = args.pop(0)
        return getattr_rec(getattr2(obj, attr), args)
    else:
        from django.core.exceptions import ObjectDoesNotExist

        try:
            model = type(obj)
            attr_name = args[0]
            attr = getattr(model, attr_name)
            value = getattr(obj, attr_name)
            field = None
            if hasattr(attr, 'field_name'):
                field = getattr(model, '_meta').get_field(attr.field_name)
            elif hasattr(attr, 'field'):
                field = attr.field

            if field:
                if hasattr(field, 'formatter') and field.formatter:
                    from djangoplus.cache import loader
                    func = loader.formatters[field.formatter]
                    if len(func.__code__.co_varnames) == 1:
                        value = func(value)
                    else:
                        value = func(value, request=obj.request, verbose_name=field.verbose_name, instance=obj)
                    return mark_safe(str(value))
                elif hasattr(field, 'display') and field.display not in (True, False, None):
                    if hasattr(obj, field.display):
                        args[0] = field.display
                        return getattr_rec(obj, args)
                elif hasattr(field, 'suffix') and field.suffix:
                    suffix = field.suffix
                    if hasattr(obj, field.suffix):
                        suffix = getattr(obj, field.suffix)
                        if callable(suffix):
                            suffix = suffix()
                    return value, suffix
                elif hasattr(obj, 'get_{}_display'.format(args[0])):
                    return getattr(obj, 'get_{}_display'.format(args[0]))()
            else:
                # it is a method decorated with @meta
                _metadata = hasattr(value, '_metadata') and getattr(value, '_metadata') or None
                if _metadata:
                    verbose_name = _metadata.get('{}:verbose_name'.format(args[0]))
                    if verbose_name:
                        value = value()
                        formatter = _metadata.get('{}:formatter'.format(args[0]))
                        if formatter:
                            from djangoplus.cache import loader
                            func = loader.formatters[formatter]
                            if len(func.__code__.co_varnames) == 1:
                                value = func(value)
                            else:
                                value = func(value, request=obj.request, verbose_name=verbose_name, instance=obj)
                            return mark_safe(str(value))

            if callable(value):
                if type(value).__name__ in ('ManyRelatedManager' or 'RelatedManager'):
                    value = value.all()
                else:
                    value = value()
            return value

        except ObjectDoesNotExist:
            return None


def get_can_execute(action):
    can_execute = []
    for group_name in action.get('can_execute_by_role') or ():
        can_execute.append(group_name)
    for group_name in action.get('can_execute_by_unit') or ():
        can_execute.append(group_name)
    for group_name in action.get('can_execute_by_organization') or ():
        can_execute.append(group_name)
    for group_name in action.get('can_execute') or ():
        can_execute.append(group_name)
    return can_execute


def should_filter_or_display(request, model, to):
    if get_metadata(to, 'role_username'):
        can_view = list(
            get_metadata(model, 'can_admin', (), iterable=True) +
            get_metadata(model, 'can_admin_by_organization', (), iterable=True) +
            get_metadata(model, 'can_admin_by_unit', (), iterable=True) +
            get_metadata(model, 'can_admin_by_role', (), iterable=True) +
            get_metadata(model, 'can_list', (), iterable=True) +
            get_metadata(model, 'can_list_by_organization', (), iterable=True) +
            get_metadata(model, 'can_list_by_unit', (), iterable=True) +
            get_metadata(model, 'can_list_by_role', (), iterable=True) +
            get_metadata(model, 'can_view', (), iterable=True) +
            get_metadata(model, 'can_view_by_organization', (), iterable=True) +
            get_metadata(model, 'can_view_by_unit', (), iterable=True) +
            get_metadata(model, 'can_view_by_role', (), iterable=True))
        if not request.user.is_superuser and can_view and not request.user.in_group(*can_view):
            return False
    return True


def find_action(model, action_name):
    from djangoplus.cache import loader
    for actions in (loader.actions, loader.class_actions):
        for action_group in actions[model]:
            for func_name, action in list(actions[model][action_group].items()):
                if action['title'] == action_name:
                    return action
    return None


def find_model_by_verbose_name(verbose_name):
    from django.apps import apps
    for model in apps.get_models():
        app_label = get_metadata(model, 'app_label')
        if not app_label.startswith('admin'):
            if get_metadata(model, 'verbose_name') == verbose_name:
                return model
    return None


def find_model_by_verbose_name_plural(verbose_name_plural):
    from django.apps import apps
    for model in apps.get_models():
        app_label = get_metadata(model, 'app_label')
        if not app_label.startswith('admin'):
            if get_metadata(model, 'verbose_name_plural') == verbose_name_plural:
                return model
    return None


def find_subset_by_title(title, model):
    from djangoplus.cache import loader
    for subset in loader.subsets[model]:
        if subset['title'] == title:
            return subset
    return None


def find_model_by_add_label(add_label):
    from django.apps import apps
    for model in apps.get_models():
        if get_metadata(model, 'add_label') == add_label:
            return model
    return None


def find_model(model, key):
    tokens = key.split('__')
    tokens.reverse()
    return find_model_recursively(model, tokens)


def find_model_recursively(cls, tokens):
    if tokens:
        token = tokens.pop()
        try:
            attr = getattr(cls, token)
            if not hasattr(attr, 'field'):
                return cls
            return find_model_recursively(attr.field.remote_field.model, tokens)
        except AttributeError as e:
            for related_object in get_metadata(cls, 'related_objects'):
                if related_object.name == token:
                    return find_model_recursively(related_object.related_model, tokens)
            raise e
    else:
        return cls


def check_condition(condition, obj):
    satisfied = True
    if obj.pk and condition:
        attr_name = condition.replace('not ', '')
        attr = getattr(obj, attr_name)
        if callable(attr):
            satisfied = attr()
        else:
            satisfied = bool(attr)
        if 'not ' in condition:
            satisfied = not satisfied
    return satisfied






# -*- coding: utf-8 -*-

from djangoplus.ui.components.utils import Timeline, QrCode, ProgressBar


def timeline(value, **kwargs):
    return Timeline(kwargs.get('request'),  kwargs.get('verbose_name'), value)


def statistics(value, **kwargs):
    value.title = kwargs.get('verbose_name')
    return value.as_table(kwargs.get('request'))


def chart(value, **kwargs):
    value.title = kwargs.get('verbose_name')
    return value.as_chart(kwargs.get('request'))


def pie_chart(value, **kwargs):
    return chart(value, **kwargs).pie()


def donut_chart(value, **kwargs):
    return chart(value, **kwargs).donut()


def box_chart(value, **kwargs):
    return chart(value, **kwargs).box()


def bar_chart(value, **kwargs):
    return chart(value, **kwargs).bar()


def horizontal_bar_chart(value, **kwargs):
    return chart(value, **kwargs).horizontal_bar()


def stack_chart(value, **kwargs):
    return chart(value, **kwargs).stack()


def horizontal_stack_chart(value, **kwargs):
    return chart(value, **kwargs).horizontal_stack()


def line_chart(value, **kwargs):
    return chart(value, **kwargs).line()


def area_chart(value, **kwargs):
    return chart(value, **kwargs).area()


def qrcode(value, **kwargs):
    request = kwargs.get('request', None)
    return QrCode(request, value)


def progress(percentual, **kwargs):
    request = kwargs.get('request', None)
    return ProgressBar(request, percentual)




# -*- coding: utf-8 -*-
import traceback
from django.apps import apps
from django.conf import settings
from django.shortcuts import render
from djangoplus.cache import loader
from django.http import HttpResponse
from djangoplus.ui import ComponentHasResponseException
from djangoplus.utils import permissions
from djangoplus.ui.components.panel import ModelPanel
from djangoplus.ui.components.navigation.breadcrumbs import httprr
from django.template import Template, Context
from djangoplus.ui.components.paginator import Paginator
from django.views.defaults import page_not_found, server_error
from django.core.exceptions import ValidationError
from django.http.response import HttpResponseForbidden
from django.contrib.contenttypes.models import ContentType
from djangoplus.utils.metadata import list_related_objects, \
    is_many_to_many, is_one_to_one, get_metadata, check_condition, is_one_to_many, getattr2, is_one_to_many_reverse
from djangoplus.ui.components.forms import factory, DEFAULT_FORM_TITLE, DEFAULT_SUBMIT_LABEL


def listt(request, app, cls, subset=None):

    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')
    title = get_metadata(_model, 'verbose_name_plural')
    subsetp = None
    list_display = None
    list_filter = None
    search_fields = None
    if subset:
        subset_func = getattr(_model.objects.get_queryset(), subset)
        can_view = subset_func._metadata['{}:can_view'.format(subset)]
        list_display = subset_func._metadata['{}:list_display'.format(subset)]
        list_filter = subset_func._metadata['{}:list_filter'.format(subset)]
        search_fields = subset_func._metadata['{}:search_fields'.format(subset)]
        title = '{} - {}'.format(title, subset_func._metadata['{}:title'.format(subset)])
    else:
        tid = request.GET.get('tid')
        subsetp = request.GET.get('tab{}'.format(tid))
        if tid and subsetp:
            subset_func = getattr(_model.objects.get_queryset(), subsetp)
            subset_title = subset_func._metadata['{}:title'.format(subsetp)]
            can_view = subset_func._metadata['{}:can_view'.format(subsetp)]
            title = '{} - {}'.format(title, subset_func._metadata['{}:title'.format(subsetp)])
            if not permissions.check_group_or_permission(request, can_view):
                return httprr(request, '/admin/login/?next={}'.format(request.get_full_path()))
        else:
            permission = '{}.list_{}'.format(app, cls)
            if not request.user.has_perm(permission):
                return httprr(request, '/admin/login/?next={}'.format(request.get_full_path()))

    qs = _model.objects.all(request.user)
    list_subsets = subset and [subset] or None

    paginator = Paginator(request, qs, title, list_subsets=list_subsets, is_list_view=True, list_display=list_display, list_filter=list_filter, search_fields=search_fields)
    paginator.process_request()



    paginator.add_actions()
    return render(request, 'default.html', locals())


def add(request, app, cls, pk=None, related_field_name=None, related_pk=None):

    if not request.user.is_authenticated:
        return httprr(request, '/admin/login/?next={}'.format(request.get_full_path()))

    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')

    obj = pk and _model.objects.all(request.user).filter(pk=pk).first() or _model()
    obj.request = request
    obj._user = request.user

    title = pk and str(obj) or get_metadata(_model, 'verbose_name')

    if related_field_name is None:

        if obj.pk:
            if not permissions.has_edit_permission(request, _model) or not permissions.can_edit(request, obj):
                return HttpResponseForbidden()
        else:
            if not permissions.has_add_permission(request, _model) or not permissions.can_add(request, obj):
                return HttpResponseForbidden()

        form = factory.get_register_form(request, obj)
        title = form.title

    elif is_one_to_many(_model, related_field_name):
        if not permissions.can_add(request, obj) and not permissions.can_edit(request, obj):
            return HttpResponseForbidden()
        form = factory.get_one_to_many_form(request, obj, related_field_name)

    elif is_many_to_many(_model, related_field_name):
        if not permissions.can_edit_field(request, obj, related_field_name):
            return HttpResponseForbidden()
        form = factory.get_many_to_many_form(request, obj, related_field_name, related_pk)

    elif is_one_to_many_reverse(_model, related_field_name):
        form = factory.get_many_to_many_reverse_form(request, obj, related_field_name)

    elif is_one_to_one(_model, related_field_name):
        if not permissions.can_add(request, obj) and not permissions.can_edit(request, obj):
            return HttpResponseForbidden()
        form = factory.get_one_to_one_form(request, obj, related_field_name, related_pk)
    else:
        # many to one
        for rel in list_related_objects(_model):
            if hasattr(rel, 'get_accessor_name'):
                if rel.get_accessor_name() in ('{}_set'.format(related_field_name), related_field_name):
                    related_queryset = rel.related_model.objects.all(request.user)
                    related_obj = related_pk and related_queryset.get(pk=related_pk) or rel.related_model()
                    related_obj.request = request
                    setattr(related_obj, rel.field.name, obj)
                    setattr(related_obj, '{}_id'.format(rel.field.name), obj.pk)
                    if related_pk:
                        if not permissions.has_edit_permission(request, rel.related_model) or not permissions.can_edit(request, related_obj):
                            return HttpResponseForbidden()
                    else:
                        if not permissions.has_add_permission(request, rel.related_model) or not permissions.can_add(request, related_obj):
                            return HttpResponseForbidden()
                    form = factory.get_many_to_one_form(request, obj, rel.get_accessor_name(), related_obj)
                    title = form.title

    if form.is_valid():
        is_editing = form.instance.pk is not None
        try:
            form.save()
            obj = form.instance
            if 'select' in request.GET:
                return HttpResponse('{}|{}|{}'.format(obj.pk, obj, request.GET['select']));
            elif related_field_name:
                message = 'Ação realizada com sucesso'
                url = '..'
            else:
                message = get_metadata(form.instance.__class__, 'add_message')
                if message and not is_editing:
                    if hasattr(obj, 'get_absolute_url'):
                        url = obj.get_absolute_url()
                    else:
                        url = '/view/{}/{}/{}/'.format(get_metadata(obj.__class__, 'app_label'), obj.__class__.__name__.lower(), obj.pk)
                else:
                    url = '..'
                if is_editing:
                    message = message or 'Atualização realizada com sucesso'
                else:
                    message = message or 'Cadastro realizado com sucesso'
            return httprr(request, url, message)
        except ValidationError as e:
            form.add_error(None, str(e.message))
    return render(request, 'default.html', locals())


def view(request, app, cls, pk, tab=None):

    if not request.user.is_authenticated:
        return httprr(request, '/admin/login/?next={}'.format(request.get_full_path()))

    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')

    obj = _model.objects.all(request.user).filter(pk=pk).first()
    obj.request = request
    obj._user = request.user

    if 'one_to_many_count' in request.GET:
        # TODO create a specific view for this purpose
        return HttpResponse(getattr2(obj, request.GET['one_to_many_count']))

    if not permissions.can_view(request, obj):
        return HttpResponseForbidden()

    title = str(obj)
    parent = request.GET.get('parent', None)
    printable = get_metadata(_model, 'pdf', False)
    panel = ModelPanel(request, obj, tab, parent, printable=printable)
    panel.process_request()

    if panel.message:
        return httprr(request, request.get_full_path(), panel.message)

    log_data = get_metadata(obj.__class__, 'log', False)
    if log_data and request.user.is_superuser and request.user.has_perm('admin.list_log'):
        url = '/log/{}/{}/'.format(app, cls)
        panel.drop_down.add_action('Visualizar Log', url, 'ajax', 'fa fa-history')

    return render(request, 'default.html', locals())


def action(request, app, cls, action_name, pk=None):

    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')

    for group in loader.actions[_model]:
        if action_name in loader.actions[_model][group]:
            break

    form_action = loader.actions[_model][group][action_name]
    action_title = form_action['title']
    action_can_execute = form_action['can_execute']
    action_condition = form_action['condition']
    action_function = form_action['function']
    action_message = 'message' in form_action and form_action['message'] or None
    action_permission = '{}.{}'.format(_model._meta.app_label, action_function.__name__)
    action_input = form_action['input']
    action_display = form_action['display']
    action_style = form_action['css']
    action_redirect = form_action['redirect_to']

    obj = pk and _model.objects.all(request.user).distinct().get(pk=pk) or _model()
    obj.request = request
    obj._user = request.user
    title = action_title
    redirect_to = None

    if check_condition(action_condition, obj) and (not action_can_execute or permissions.check_group_or_permission(request, action_permission)):
        f_return = None
        func = getattr(_model, action_function.__name__, action_function)
        form = factory.get_action_form(request, obj, form_action)

        if func.__code__.co_argcount > 1 or action_input:
            if form.is_valid():
                if 'instance' in form.fields:
                    obj = form.cleaned_data['instance']
                func = getattr(obj, action_function.__name__, action_function)
                params = []
                for param in func.__code__.co_varnames[1:func.__code__.co_argcount]:
                    if param in form.cleaned_data:
                        params.append(form.cleaned_data[param])
                try:
                    f_return = func(*params)
                    if not action_redirect:
                        if func.__code__.co_argcount > 1 or action_display:
                            redirect_to = '..'
                        else:
                            redirect_to = '.'
                    else:
                        redirect_to = Template(action_redirect).render(Context({'self': obj}))
                except ValidationError as e:
                    form.add_error(None, str(e.message))
        else:
            try:
                if form.fields and form.is_valid() or not form.fields:
                    if 'instance' in form.fields:
                        obj = form.cleaned_data['instance']
                    func = getattr(obj, action_function.__name__, action_function)
                    f_return = func()
                    if not action_redirect:
                        if func.__code__.co_argcount > 1 or action_display:
                            redirect_to = '..'
                        else:
                            redirect_to = '.'
                    else:
                        redirect_to = Template(action_redirect).render(Context({'self': obj}))
            except ValidationError as e:
                if form.fields:
                    form.add_error(None, str(e.message))
                return httprr(request, '.', e.message, error=True)

        if f_return:
            if 'pdf' in action_style:
                request.GET._mutable = True
                request.GET['pdf'] = 1
                request.GET._mutable = False
                from datetime import datetime
                from djangoplus.admin.models import Settings
                from djangoplus.utils.http import PdfResponse
                from django.template.loader import render_to_string
                app_settings = Settings.default()
                f_return['logo'] = app_settings.logo_pdf and app_settings.logo_pdf or app_settings.logo
                f_return['project_name'] = app_settings.initials
                f_return['project_description'] = app_settings.name
                f_return['today'] = datetime.now()
                template_list = ['{}.html'.format(action_function.__name__), 'report.html']
                landscape = 'landscape' in action_style
                return PdfResponse(render_to_string(template_list, f_return, request=request), landscape=landscape)
        elif redirect_to:
            return httprr(request, redirect_to, action_message)

        if form.title == DEFAULT_FORM_TITLE:
            form.title = action_title
        if form.submit_label == DEFAULT_SUBMIT_LABEL:
            form.submit_label = action_title
        return render(request, 'default.html', locals())
    else:
        return HttpResponseForbidden()


def delete(request, app, cls, pk, related_field_name=None, related_pk=None):

    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')

    obj = _model.objects.all(request.user).get(pk=pk)
    obj._request = request
    obj._user = request.user

    if permissions.can_delete(request, obj):
        if related_field_name:
            getattr(obj, related_field_name).remove(related_pk)
            return httprr(request, '..', 'Removido com sucesso')
        else:
            title = 'Excluir {}'.format(str(obj))
            form = factory.get_delete_form(request, obj)
            if form.is_valid():
                obj.delete()
                return httprr(request, '..', 'Exclusão realizada com sucesso.')
            return render(request, 'delete.html', locals())
    else:
        return HttpResponseForbidden()


def log(request, app, cls, pk=None):
    try:
        _model = apps.get_model(app, cls)
    except LookupError as e:
        return page_not_found(request, e, 'error404.html')

    if pk:
        obj = _model.objects.get(pk=pk)
        qs = obj.get_logs()
        title = 'Log - {}'.format(obj)
    else:
        content_type = ContentType.objects.get_for_model(_model)
        qs = content_type.log_set.all()
        title = 'Logs - {}'.format(get_metadata(_model, 'verbose_name_plural'))

    paginator = Paginator(request, qs, 'Log')
    return render(request, 'default.html', locals())


def dispatcher(request, app, view_name, params):

    params = params.split('/')[0:-1]

    full_app_name = settings.APP_MAPPING.get(app, app)
    fromlist = full_app_name.split('.')

    try:
        views = __import__('{}.views'.format(full_app_name), fromlist=list(map(str, fromlist)))
        func = getattr(views, view_name)
    except ComponentHasResponseException as e:
        raise e
    except (ImportError, TypeError, AttributeError) as e:
        traceback.print_exc()
        return page_not_found(request, e, 'error404.html')

    try:
        return func(request, *params)
    except ComponentHasResponseException as e:
        raise e
    except Exception as e:
        print(e)
        traceback.print_exc()
        #return server_error(request, 'error500.html')
        raise e





# -*- coding: utf-8 -*-
import time
import djangoplus
from django.core.management.base import BaseCommand
from djangoplus.tutorial.tools import EditorSimulator, VideoRecorder

BASE_DIR = djangoplus.__path__[0]


def test1():
    editor = EditorSimulator('{}/tutorial/tools.py'.format(BASE_DIR))
    editor.proccess_lines()
    editor.write('Breno! =)', 11)
    editor.write('# OK', 20)
    editor.write(', muito massa', 11, 5)
    editor.write('# primeira linha', 1)
    editor.write('# linha 40', 40)
    editor.save('/tmp/output.txt')
    editor.close()


def test2():
    editor = EditorSimulator('{}/tutorial/files/sample.txt'.format(BASE_DIR))
    editor.simulate(3, 5, pause=False)
    time.sleep(3)
    editor.simulate(8, 9, pause=True)


def test3():
    editor = EditorSimulator()
    editor.write('linha 1', 1)
    editor.write('linha 2', 2)
    editor.write('complemento da linha 2', 2, 8)
    editor.write('linha 3', 3)
    editor.write('linha 4', 4)
    for i in range(5, 25):
        editor.write('linha {}'.format(i), i)
    editor.write('Nova linha 2', 2)
    editor.write('Nova linha 20', 20)
    editor.write('Nova linha 13', 13)
    editor.write('Nova linha 14', 14)
    editor.write('Nova linha 15', 15)
    editor.close()


def test4():
    editor = EditorSimulator('{}/tutorial/tools.py'.format(BASE_DIR), slow=False)
    editor.simulate()
    editor.save('/tmp/output.txt')


class Command(BaseCommand):

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('file_path', default=None)
        parser.add_argument('start', type=float, nargs='?', default=None)
        parser.add_argument('end', type=float, nargs='?', default=None)
        parser.add_argument('--record', action='store_true', dest='record', default=False)
        parser.add_argument('--fast', action='store_true', dest='fast', default=False)

    def handle(self, *args, **options):
        recorder = VideoRecorder()
        file_path = options.pop('file_path')
        start = options.pop('start')
        end = options.pop('end')
        record = options.pop('record', False)
        fast = options.pop('fast', False)

        if record:
            recorder.start()
        try:
            if file_path == '1':
                test1()
            elif file_path == '2':
                test2()
            elif file_path == '3':
                test3()
            elif file_path == '4':
                test4()
            else:
                editor = EditorSimulator(file_path, slow=not fast)
                editor.simulate(start_step=start, end_step=end)
        finally:
            if record:
                recorder.stop(file_path.split('/')[-1])




# -*- coding: utf-8 -*-
from tempfile import mktemp
from django.template.loader import render_to_string
from django.http import HttpResponse
from django.conf import settings
import tempfile
import json
import os
import datetime
from djangoplus.utils.formatter import to_ascii


def mobile(request):
    for device in ['iphone', 'android']:
        if device in request.META.get('HTTP_USER_AGENT', '').lower():
            return True
    return False


class XlsResponse(HttpResponse):
    def __init__(self, data, name='Listagem'):
        import xlwt
        wb = xlwt.Workbook(encoding='iso8859-1')
        for title, rows in data:
            sheet = wb.add_sheet(title)
            for row_idx, row in enumerate(rows):
                for col_idx, label in enumerate(row):
                    sheet.write(row_idx, col_idx, label=label)
        file_name = mktemp()
        wb.save(file_name)
        HttpResponse.__init__(self, content=open(file_name, 'rb').read(), content_type='application/vnd.ms-excel')
        self['Content-Disposition'] = 'attachment; filename={}.xls'.format(to_ascii(name))


class CsvResponse(HttpResponse):
    def __init__(self, rows, name='Listagem'):
        import csv
        file_name = mktemp()
        with open(file_name, 'w', encoding='iso8859-1') as output:
            writer = csv.writer(output)
            for row in rows:
                writer.writerow([col for col in row]) # .encode('iso8859-1')
        HttpResponse.__init__(self, content=open(file_name, 'r').read(), content_type='application/csv')
        self['Content-Disposition'] = 'attachment; filename={}.xls'.format(to_ascii(name))


class ZipResponse(HttpResponse):
    def __init__(self, file_path):
        content = open(file_path)
        HttpResponse.__init__(self, content=content, content_type='application/zip')
        self['Content-Disposition'] = 'attachment; filename={}'.format(to_ascii(file_path.split(os.sep)[-1]))


class JsonResponse(HttpResponse):
    def __init__(self, data):
        content = json.dumps(data)
        HttpResponse.__init__(self, content=content, content_type='application/json')


class PdfResponse(HttpResponse):

    def __init__(self, html, landscape=False):
        from xhtml2pdf import pisa

        def link_callback(uri, rel):
            s = '{}/{}'.format(settings.MEDIA_ROOT, uri.replace('/media', ''))
            return s

        tmp = tempfile.NamedTemporaryFile(mode='w+b', delete=False)
        file_name = tmp.name
        if landscape:
            html = html.replace('a4 portrait', 'a4 landscape')
            html = html.replace('logo_if_portrait', 'logo_if_landscape')
        out = pisa.CreatePDF(html, tmp, link_callback=link_callback)
        out.dest.close()
        tmp = open(file_name, "rb")
        str_bytes = tmp.read()
        os.unlink(file_name)
        HttpResponse.__init__(self, str_bytes, content_type='application/pdf')


class ReportResponse(PdfResponse):
    def __init__(self, title, request, objects, landscape=False, template='report.html'):
        from djangoplus.admin.models import Settings
        app_settings = Settings.default()
        logo = app_settings.logo_pdf and app_settings.logo_pdf or app_settings.logo
        project_name = app_settings.initials
        project_description = app_settings.name
        unit_or_organization = request.session.get('scope')
        context = dict(objects=objects, title=title, today=datetime.date.today(), logo=logo,
                       project_name=project_name, project_description=project_description,
                       unit_or_organization=unit_or_organization)
        html = render_to_string([template], context, request=request)
        super(ReportResponse, self).__init__(html, landscape)


class ComponentResponse(HttpResponse):
    def __init__(self, component):
        super(ComponentResponse, self).__init__(
            render_to_string('default.html', context=dict(paginator=component), request=component.request)
        )


# -*- coding: utf-8 -*-

import os
import datetime
from fabric.api import *
from os import path
import time
EMPTY_TEST_FILE_CONTENT = '''# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from djangoplus._test_admin.models import User
from djangoplus.test import TestCase
from django.conf import settings


class AppTestCase(TestCase):

    def test_app(self):

        User.objects.create_superuser('_test_admin', None, settings.DEFAULT_PASSWORD)

        self.login('_test_admin', settings.DEFAULT_PASSWORD)
'''


DOCKER_FILE_CONTENT = '''FROM {}
ENV DEBIAN_FRONTEND=noninteractive
RUN apt-get update
RUN apt-get -y install python3 python3-pip build-essential python3-dev libfreetype6-dev python3-cffi libtiff5-dev liblcms2-dev libwebp-dev tk8.6-dev libjpeg-dev ssh openssh-server dnsutils curl vim git
RUN apt-get -y install chrpath libssl-dev libxft-dev libfreetype6 libfreetype6-dev libfontconfig1 libfontconfig1-dev wget

RUN apt-get -y install libgtk-3-dev
RUN wget https://ftp.mozilla.org/pub/firefox/releases/60.0b3/linux-x86_64/en-US/firefox-60.0b3.tar.bz2
RUN tar xvjf firefox-60.0b3.tar.bz2
RUN ln -s /firefox/firefox /usr/local/bin/firefox
RUN wget https://github.com/mozilla/geckodriver/releases/download/v0.20.0/geckodriver-v0.20.0-linux64.tar.gz
RUN gunzip geckodriver-v0.20.0-linux64.tar.gz
RUN tar -xvf geckodriver-v0.20.0-linux64.tar
RUN mv geckodriver /usr/local/bin/

ENV LC_ALL=C.UTF-8
ENV LANG=C.UTF-8
RUN export LANG=C.UTF-8

RUN ln -s /usr/bin/pip3 /usr/bin/pip
RUN ln -sfn /usr/bin/python3 /usr/bin/python

RUN pip install --upgrade pip
'''

PROJECTS = [
    ('companies', 'git@djangoplus.net:companies.git'),
    ('blackpoint', 'git@bitbucket.org:brenokcc/blackpoint.git'),
    ('emprestimos', 'git@djangoplus.net:emprestimos.git'),
    ('financeiro', 'git@bitbucket.org:brenokcc/financeiro.git'),
    ('formulacao', 'git@bitbucket.org:brenokcc/formulacao.git'),
    ('gouveia', 'git@bitbucket.org:brenokcc/gouveia.git'),
    ('petshop', 'git@bitbucket.org/brenokcc/petshop.git'),
    ('loja', 'git@bitbucket.org/brenokcc/loja.git'),
    ('biblioteca', 'git@bitbucket.org/brenokcc/biblioteca.git'),
    ('gerifes', 'git@bitbucket.org/brenokcc/gerifes.git'),
    ('simop', 'git@bitbucket.org/brenokcc/simop.git'),
    ('fabrica', 'git@bitbucket.org:brenokcc/fabrica.git'),
    ('abstract', 'git@bitbucket.org:brenokcc/abstract.git'),
]


def _test_startpoject():
    if path.exists('/tmp/xxx'):
        local('rm -r /tmp/xxx')
    with lcd('/tmp/'):
        local('startproject xxx')
        with lcd('/tmp/xxx'):
            local('python manage.py test')
        with lcd('/tmp'):
            local('rm -r /tmp/xxx')


def _test_admin():
    local('python manage.py test djangoplus.admin.tests.AdminTestCase')


def _test_projects():
    from subprocess import Popen, PIPE
    paths = []
    start = datetime.datetime.now()
    os.environ['PYTHONIOENCODING'] = 'utf-8'
    for project_name, project_url in PROJECTS:
        if path.exists('/Users/breno'):
            base_path = '/Users/breno/Documents/Workspace'
            if project_name in ('petshop', 'loja', 'biblioteca'):
                base_path = path.join(base_path, 'djangoplus/djangoplus-demos')
            project_path = path.join(base_path, project_name)
        else:
            project_path = path.join('/tmp', project_name)
            if not path.exists(project_path):
                local('git clone {} {}'.format(project_url, project_path))
            with lcd(project_path):
                local('git pull origin master')
        paths.append(project_path)

    running_procs = []
    for project_path in paths:
        project_name = project_path.split('/')[-1]
        print('Testing {}'.format(project_name))
        with lcd(project_path):
            local('python manage.py test')
        # proc = Popen(['python', 'manage.py', 'test'], cwd=project_path, stderr=PIPE, stdout=PIPE)
        # proc.project_name = project_name
        # running_procs.append(proc)

    while running_procs:
        for proc in running_procs:
            retcode = proc.poll()
            if retcode is not None:
                running_procs.remove(proc)
                if retcode != 0 and retcode != -9:
                    print('An error was found while executing "{}"!'.format(proc.project_name))
                    print(proc.stderr.read())
                    for uncessary_proc in running_procs:
                        print('Killing execution for "{}"...'.format(uncessary_proc.project_name))
                        uncessary_proc.kill()
            else:
                time.sleep(5)
                continue
    end = datetime.datetime.now()
    print('Tests executed in "{}" seconds!!!'.format((end-start).seconds))


def _test_testcases_generation():
    test_file_path = '{}/emprestimos/emprestimos/tests.py'.format('/Users/breno/Documents/Workspace')
    test_file_content = open(test_file_path).read()
    open(test_file_path, 'w').write(EMPTY_TEST_FILE_CONTENT)
    with lcd('{}/emprestimos'.format('/Users/breno/Documents/Workspace')):
        local('python manage.py test --add')
    print(open(test_file_path).read())
    open(test_file_path, 'w').write(test_file_content)


def _test_so_installation(so):
    docker_file = open('/tmp/Dockerfile', 'w')
    docker_file.write(DOCKER_FILE_CONTENT.format(so))
    docker_file.close()
    local('docker build -t djangoplus-{} /tmp'.format(so))
    local('docker run djangoplus-{} pip install djangoplus && startproject xyz && cd xyz && python manage.py test djangoplus.admin.tests.AdminTestCase'.format(so))


def _test_deploy():
    pass


def test(scope=''):
    if scope in ('startproject', 'all'):
        _test_startpoject()
    if scope in ('admin', 'all'):
        _test_admin()
    elif scope in ('implementation', 'all'):
        _test_projects()
    elif scope in ('installation', 'all'):
        _test_so_installation('debian')
        _test_so_installation('ubuntu')
    elif scope in ('deploy', 'all'):
        _test_deploy()
    else:
        print('Available parameters: startproject, admin, implementation, installation, deploy')




# -*- coding: utf-8 -*-
from django.conf import settings
from django.core.management.base import BaseCommand


class Command(BaseCommand):

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        parser.add_argument('--continue', action='store_true', dest='continue', default=False,
                            help='Will not create a new project. The existing one will be used.')

    def handle(self, *args, **options):
        fromlist = [settings.PROJECT_NAME, 'tutorial', 'AppTutorial']
        module = __import__('{}.tutorial'.format(settings.PROJECT_NAME), fromlist=fromlist)
        tutorial = module.AppTutorial(settings.BASE_DIR)
        create_project = not options.pop('continue', False)
        tutorial.start(create_project)



# -*- coding: utf-8 -*-
from decimal import Decimal
from djangoplus.ui import RequestComponent
from django.utils.text import slugify
from djangoplus.utils import permissions
from django.template.loader import render_to_string
from djangoplus.ui.components.paginator import Paginator
from djangoplus.ui.components.navigation.dropdown import ModelDropDown, GroupDropDown
from djangoplus.utils.metadata import get_metadata, get_fieldsets, find_field_by_name, get_fiendly_name, \
    check_condition, is_one_to_one, is_many_to_one, should_filter_or_display


class Panel(RequestComponent):

    def __init__(self, request, title=None, text=None, icon=None):
        super(Panel, self).__init__(title, request)
        self.title = title
        self.text = text
        self.labels = []
        self.icon = icon


class ModelPanel(RequestComponent):
    def __init__(self, request, obj, current_tab=None, parent=None, fieldsets=None, complete=True, readonly=False, printable=True):

        super(ModelPanel, self).__init__(obj.pk, request)

        self.obj = obj
        self.title = obj.pk and str(obj) or get_metadata(type(obj), 'verbose_name')
        self.id = self.title
        self.tabs = []
        self.current_tab = current_tab
        self.message = None
        self.complete = complete
        self.readonly = readonly
        self.printable = printable
        self.drop_down = None
        fieldsets = fieldsets or get_metadata(type(obj), 'view_fieldsets', [])
        if not fieldsets:
            fieldsets = get_fieldsets(type(obj))

        if self.complete:
            self.drop_down = ModelDropDown(self.request, type(self.obj))
            self.drop_down.add_actions(self.obj, fieldset_title='')
            if self.printable:
                self.drop_down.add_action('Imprimir', url='?pdf={}&pk='.format(self.id), css='ajax', icon='fa-print', category='Imprimir')
        else:
            self.drop_down = GroupDropDown(self.request)

        self.fieldsets_with_tab_name = []
        self.fieldsets_without_tab_name = []

        model = type(self.obj)
        obj.as_pdf = self.as_pdf

        for fieldset in fieldsets:
            title, info = fieldset
            tab_name = None

            drop_down = ModelDropDown(self.request, model)
            fieldset_actions = info.get('actions', [])
            fieldset_image = info.get('image')

            if fieldset_actions:
                drop_down.add_actions(self.obj, fieldset_title=title)

            if 'condition' in fieldset[1]:
                condition = fieldset[1]['condition']
                self.obj.request = self.request
                if not check_condition(condition, self.obj):
                    continue

            if '::' in title:
                tab_name, title = title.split('::')
                url = '/view/{}/{}/{}/{}/'.format(get_metadata(model, 'app_label'), model.__name__.lower(), self.obj.pk, slugify(tab_name))
                tab = (tab_name, url)
                if not self.tabs and not self.current_tab:
                    self.current_tab = slugify(tab_name)
                if tab not in self.tabs:
                    self.tabs.append(tab)

            if not tab_name or slugify(tab_name) == self.current_tab or self.as_pdf:

                fieldset_dict = dict(title=title or 'Dados Gerais', tab_name=tab_name, fields=[], paginators=[], drop_down=drop_down, image=None)
                relations = list(fieldset[1].get('relations', []))
                inlines = list(fieldset[1].get('inlines', []))

                if tab_name or self.as_pdf:
                    self.fieldsets_with_tab_name.append(fieldset_dict)
                else:
                    self.fieldsets_without_tab_name.append(fieldset_dict)

                if 'can_view' in fieldset[1]:
                    can_view = fieldset[1]['can_view']
                    if not permissions.check_group_or_permission(self.request, can_view):
                        continue

                if 'image' in fieldset[1]:
                    fieldset_dict['image'] = fieldset[1]['image']

                if 'fields' in fieldset[1]:
                    for name_or_tuple in fieldset[1]['fields']:

                        if not type(name_or_tuple) == tuple:
                            name_or_tuple = (name_or_tuple,)
                        attr_names = []

                        for attr_name in name_or_tuple:
                            if attr_name != parent and attr_name != fieldset_image:
                                attr = getattr(model, attr_name)
                                field = None
                                if hasattr(attr, 'field_name'):
                                    field = getattr(model, '_meta').get_field(attr.field_name)
                                elif hasattr(attr, 'field'):
                                    field = attr.field
                                if not field or not hasattr(field, 'display') or field.display:
                                    verbose_name, lookup, sortable, to = get_fiendly_name(model, attr_name, as_tuple=True)
                                    if to and not should_filter_or_display(self.request, model, to):
                                        continue
                                    attr_names.append(dict(verbose_name=verbose_name, name=attr_name))
                        if attr_names:
                            fieldset_dict['fields'].append(attr_names)

                if self.complete:
                    from djangoplus.utils.relations import Relation
                    for relation_name in relations + inlines:
                        component = Relation(self.obj, relation_name).get_component(self.request, self.as_pdf)
                        fieldset_dict['paginators'].append(component)

                else:
                    for relation_name in relations + inlines:
                        if relation_name in [field.name for field in get_metadata(model, 'get_fields')]:
                            relation_field = find_field_by_name(model, relation_name)
                            if is_one_to_one(model, relation_name) or is_many_to_one(model, relation_name):
                                fieldset_dict['fields'].append(
                                    [dict(verbose_name=relation_field.verbose_name, name=relation_name)]
                                )

                if 'extra' in fieldset[1]:
                    fieldset_dict['extra'] = []
                    for info in fieldset[1]['extra']:
                        fieldset_dict['extra'].append(info)

    def get_active_fieldsets(self):
        return self.fieldsets_without_tab_name + self.fieldsets_with_tab_name


class IconPanel(RequestComponent):

    def __init__(self, request, title, icon=None):
        super(IconPanel, self).__init__('iconpanel', request)
        self.title = title
        self.icon = icon
        self.items = []

    def add_item(self, label, url, icon=None, css='ajax'):
        item = dict(label=label, url=url, css=css, icon=icon or self.icon)
        self.items.append(item)


class ShortcutPanel(RequestComponent):

    def __init__(self, request):
        from djangoplus.cache import loader
        super(ShortcutPanel, self).__init__('shortcutpanel', request)
        self.items = []

        for model, add_shortcut in loader.icon_panel_models:

            if type(add_shortcut) == bool:
                add_model = add_shortcut
            else:
                if not type(add_shortcut) in (list, tuple):
                    add_shortcut = add_shortcut,
                add_model = request.user.in_group(*add_shortcut)

            if add_model or request.user.is_superuser:
                if permissions.has_add_permission(request, model):
                    self.add_model(model)
        for item in loader.views:
            if item['add_shortcut']:
                if permissions.check_group_or_permission(request, item['can_view']):
                    self.add(item['icon'], item['title'], None, item['url'], item['can_view'], item['style'])

    def add(self, icon, description, count=None, url=None, perm_or_group=None, style='ajax'):
        if permissions.check_group_or_permission(self.request, perm_or_group):
            item = dict(icon=icon, description=description, count=count, url=url or '#', style=style)
            self.items.append(item)

    def add_model(self, model):
        icon = get_metadata(model, 'icon')
        prefix = get_metadata(model, 'verbose_female') and 'Nova' or 'Novo'
        description = '{} {}'.format(prefix, get_metadata(model, 'verbose_name'))
        description = get_metadata(model, 'add_label', description)
        url = '/add/{}/{}/'.format(get_metadata(model, 'app_label'), model.__name__.lower())
        permission = '{}.list_{}'.format(get_metadata(model, 'app_label'), model.__name__.lower())
        self.add(icon, description, None, url, permission)

    def add_models(self, *models):
        for model in models:
            self.add_model(model)


class CardPanel(RequestComponent):

    class Media:
        pass
        # css = {'all': ('/static/css/cardpanel.css',)}

    def __init__(self, request):
        super(CardPanel, self).__init__('cardpanel', request)

        self.items = []

        from djangoplus.cache import loader
        for model, list_shortcut in loader.card_panel_models:

            if type(list_shortcut) == bool:
                add_model = list_shortcut
            else:
                if not type(list_shortcut) in (list, tuple):
                    list_shortcut = list_shortcut,
                add_model = request.user.in_group(*list_shortcut)

            if add_model or request.user.is_superuser:
                if permissions.has_list_permission(request, model):
                    self.add_model(model)

                if model in loader.subsets:
                    icon = get_metadata(model, 'icon')
                    title = get_metadata(model, 'verbose_name_plural')
                    app_label = get_metadata(model, 'app_label')
                    model_name = model.__name__.lower()
                    for item in loader.subsets[model]:
                        can_view = item['can_view']
                        # TODO False
                        if False and permissions.check_group_or_permission(self.request, can_view):
                            attr_name = item['function'].__func__.__name__
                            qs = model.objects.all(self.request.user)
                            qs = getattr(qs, attr_name)()
                            count = qs.count()
                            if count:
                                url = '/list/{}/{}/{}/'.format(app_label, model_name, attr_name)
                                self.add(icon, title, count, url, '', None, item['title'])

        for item in loader.views:
            if False:  # TODO False
                if permissions.check_group_or_permission(request, item['can_view']):
                    self.add(item['icon'], item['menu'], None, item['url'], '', item['can_view'], item['style'])

    def add(self, icon, title, count=None, url=None, css='', perm_or_group=None, description=''):
        if permissions.check_group_or_permission(self.request, perm_or_group):
            item = dict(icon=icon, title=title, count=count, url=url, css=css, description=description)
            self.items.append(item)

    def add_model(self, model):
        self.add_models(model)

    def add_models(self, *models):
        for model in models:
            icon = get_metadata(model, 'icon')
            title = get_metadata(model, 'verbose_name_plural')
            app_label = get_metadata(model, 'app_label')
            model_name = model.__name__.lower()
            url = '/list/{}/{}/'.format(app_label, model_name)
            permission = '{}.list_{}'.format(app_label, model_name)
            self.add(icon, title, None, url, '', permission)


class DashboardPanel(RequestComponent):

    def __init__(self, request):

        super(DashboardPanel, self).__init__('dashboard', request)
        self.top = []
        self.center = []
        self.left = []
        self.right = []
        self.bottom = []

        from djangoplus.cache import loader
        for model in loader.subsets:
            icon = get_metadata(model, 'icon', 'fa-bell-o')
            title = get_metadata(model, 'verbose_name_plural')
            app_label = get_metadata(model, 'app_label')
            model_name = model.__name__.lower()
            for item in loader.subsets[model]:
                description = item['title']
                notify = item['notify']
                if notify is True or notify and permissions.check_group_or_permission(request, notify):
                    attr_name = item['function'].__func__.__name__
                    qs = model.objects.all(request.user)
                    qs = getattr(qs, attr_name)()
                    count = qs.count()
                    if count:
                        url = '/list/{}/{}/{}/'.format(app_label, model_name, attr_name)
                        notification_panel = NotificationPanel(request, title, count, url, description, icon)
                        self.right.append(notification_panel)

        icon_panel = ShortcutPanel(request)
        card_panel = CardPanel(request)

        self.top.append(icon_panel)
        self.center.append(card_panel)

        for item in loader.subset_widgets:

            model = item['model']
            title = item['title']
            function = item['function']
            dashboard = item['dashboard']
            formatter = item['formatter']
            list_display = item.get('list_display')
            link = item['link']

            l = []
            if type(dashboard) == dict:
                for position, group_names in list(dashboard.items()):
                    group_names = type(group_names) == tuple and group_names or (group_names,)
                    l.append((position, group_names))
            else:
                l.append((dashboard, item['can_view']))

            for position, group_names in l:
                if permissions.check_group_or_permission(request, group_names, ignore_superuser=True):
                    qs = model.objects.all(request.user)
                    f_return = getattr(qs, function)()
                    html = ''

                    if type(f_return) in (int, Decimal):
                        verbose_name = get_metadata(model, 'verbose_name_plural')
                        icon = get_metadata(model, 'icon')
                        panel = NumberPanel(request, verbose_name, f_return, title, icon)
                        html = str(panel)

                    if type(f_return).__name__ == 'QueryStatistics' and not formatter:
                        formatter = 'statistics'

                    if formatter:
                        func = loader.formatters[formatter]
                        if len(func.__code__.co_varnames) == 1:
                            html = str(func(f_return))
                        else:
                            html = str(func(f_return, request=self.request, verbose_name=title))
                    elif hasattr(f_return, 'model'):
                        compact = position in ('left', 'right')
                        app_label = get_metadata(model, 'app_label')
                        model_name = model.__name__.lower()
                        verbose_name_plural = get_metadata(model, 'verbose_name_plural')
                        if link:
                            title = '{} {}'.format(verbose_name_plural, title)
                        url = '/list/{}/{}/{}/'.format(app_label, model_name, function)
                        paginator = Paginator(self.request, f_return, title, readonly=compact, list_display=list_display, list_filter=(), search_fields=(), list_subsets=[function], url=link and url or None)
                        if compact:
                            paginator.column_names = paginator.column_names[0:1]
                        html = str(paginator)

                    if position == 'top':
                        self.top.append(html)
                    elif position == 'center':
                        self.center.append(html)
                    elif position == 'left':
                        self.left.append(html)
                    elif position == 'right':
                        self.right.append(html)
                    elif position == 'bottom':
                        self.bottom.append(html)

        for item in loader.widgets:
            if permissions.check_group_or_permission(request, item['can_view'], ignore_superuser=True):
                function = item['function']
                position = item['position']
                f_return = function(request)
                html = render_to_string(['{}.html'.format(function.__name__), 'dashboard.html'], f_return, request)
                if position == 'top':
                    self.top.append(html)
                elif position == 'center':
                    self.center.append(html)
                elif position == 'left':
                    self.left.append(html)
                elif position == 'right':
                    self.right.append(html)
                elif position == 'bottom':
                    self.bottom.append(html)


class NumberPanel(RequestComponent):

    def __init__(self, request, title, number, description, icon=None):
        super(NumberPanel, self).__init__(title, request)
        self.title = title
        self.number = number
        self.description = description
        self.icon = icon or 'fa-comment-o'


class NotificationPanel(RequestComponent):

    def __init__(self, request, title, count, url, description, icon=None):
        super(NotificationPanel, self).__init__(title, request)
        self.title = title
        self.count = count
        self.url = url
        self.description = description
        self.icon = icon or 'fa-bell-o'



# -*- coding: utf-8 -*-

# Generated by Django 1.11 on 2018-01-06 19:21


from django.db import migrations
import djangoplus.db.models.fields


class Migration(migrations.Migration):

    dependencies = [
        ('admin', '0013_auto_20170910_1543'),
    ]

    operations = [
        migrations.AlterField(
            model_name='organization',
            name='ascii',
            field=djangoplus.db.models.fields.SearchField(blank=True, default=b'', editable=False),
        ),
        migrations.AlterField(
            model_name='unit',
            name='ascii',
            field=djangoplus.db.models.fields.SearchField(blank=True, default=b'', editable=False),
        ),
    ]



# -*- coding: utf-8 -*-
import time
from django.core.management.base import BaseCommand
from djangoplus.tutorial.tools import VideoRecorder, Subtitle
from djangoplus.utils.webdriver import Browser


class Command(BaseCommand):

    def handle(self, *args, **options):
        subtitle = Subtitle()
        recorder = VideoRecorder()
        browser = Browser('http://google.com.br')
        browser.open('/')
        recorder.start()
        subtitle.display('This is only a test', 3)
        recorder.stop('Test')
        browser.close()
        browser.service.stop()
